"""Operations on the kernel object."""
from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import itertools
import logging
import sys
from functools import reduce
from sys import intern
from typing import TYPE_CHECKING, AbstractSet, Mapping, Sequence

import numpy as np

import islpy as isl
from islpy import dim_type
from pytools import memoize_on_first_arg, natsorted

from loopy.diagnostic import LoopyError, warn_with_kernel
from loopy.kernel import LoopKernel
from loopy.kernel.function_interface import CallableKernel
from loopy.kernel.instruction import (
    CInstruction,
    MultiAssignmentBase,
    _DataObliviousInstruction,
)
from loopy.symbolic import CombineMapper
from loopy.translation_unit import TranslationUnit, TUnitOrKernelT, for_each_kernel


if TYPE_CHECKING:
    from pytools.tag import Tag

    from loopy.types import ToLoopyTypeConvertible


logger = logging.getLogger(__name__)


# {{{ add and infer argument dtypes

def add_dtypes(
            kernel: TUnitOrKernelT,
            dtype_dict: Mapping[str, ToLoopyTypeConvertible],
        ) -> TUnitOrKernelT:
    """Specify remaining unspecified argument/temporary variable types.

    :arg dtype_dict: a mapping from variable names to :class:`numpy.dtype`
        instances
    """
    if isinstance(kernel, TranslationUnit):
        t_unit = kernel
        del kernel
        kernel_names = [clbl.subkernel.name for clbl in
                t_unit.callables_table.values() if isinstance(clbl,
                    CallableKernel)]
        if len(kernel_names) != 1:
            raise LoopyError("add_dtypes may not take a TranslationUnit with more"
                    " than one callable kernels. Please provide individual kernels"
                    " instead.")

        kernel_name, = kernel_names

        return t_unit.with_kernel(
                add_dtypes(t_unit[kernel_name], dtype_dict))

    assert isinstance(kernel, LoopKernel)

    processed_dtype_dict = {}

    for k, v in dtype_dict.items():
        for subkey in k.split(","):
            subkey = subkey.strip()
            if subkey:
                processed_dtype_dict[subkey] = v

    dtype_dict_remainder, new_args, new_temp_vars = _add_dtypes(
            kernel, processed_dtype_dict)

    if dtype_dict_remainder:
        raise RuntimeError("unused argument dtypes: %s"
                % ", ".join(dtype_dict_remainder))

    return kernel.copy(args=new_args, temporary_variables=new_temp_vars)


def _add_dtypes_overdetermined(kernel, dtype_dict):
    _dtype_dict_remainder, new_args, new_temp_vars = _add_dtypes(kernel, dtype_dict)
    # do not throw error for unused args
    return kernel.copy(args=new_args, temporary_variables=new_temp_vars)


def _add_dtypes(kernel, dtype_dict):
    dtype_dict = dtype_dict.copy()
    new_args = []

    from loopy.types import to_loopy_type
    for arg in kernel.args:
        new_dtype = dtype_dict.pop(arg.name, None)
        if new_dtype is not None:
            new_dtype = to_loopy_type(new_dtype)
            if arg.dtype is not None and arg.dtype != new_dtype:
                raise RuntimeError(
                        "argument '%s' already has a different dtype "
                        "(existing: %s, new: %s)"
                        % (arg.name, arg.dtype, new_dtype))
            arg = arg.copy(dtype=new_dtype)

        new_args.append(arg)

    new_temp_vars = kernel.temporary_variables.copy()

    import loopy as lp
    for tv_name in kernel.temporary_variables:
        new_dtype = dtype_dict.pop(tv_name, None)
        if new_dtype is not None:
            new_dtype = np.dtype(new_dtype)
            tv = new_temp_vars[tv_name]
            if (tv.dtype is not None and tv.dtype is not lp.auto) \
                    and tv.dtype != new_dtype:
                raise RuntimeError(
                        "temporary variable '%s' already has a different dtype "
                        "(existing: %s, new: %s)"
                        % (tv_name, tv.dtype, new_dtype))

            new_temp_vars[tv_name] = tv.copy(dtype=new_dtype)

    return dtype_dict, new_args, new_temp_vars


def get_arguments_with_incomplete_dtype(kernel):
    return [arg.name for arg in kernel.args
            if arg.dtype is None]


def add_and_infer_dtypes(prog, dtype_dict, expect_completion=False,
        kernel_name=None):
    assert isinstance(prog, TranslationUnit)
    if kernel_name is None:
        kernel_names = [clbl.subkernel.name for clbl in
                prog.callables_table.values() if isinstance(clbl,
                    CallableKernel)]
        if len(kernel_names) != 1:
            raise LoopyError("Provide 'kernel_name' argument.")

        kernel_name, = kernel_names

    prog = prog.with_kernel(add_dtypes(prog[kernel_name], dtype_dict))

    from loopy.type_inference import infer_unknown_types
    return infer_unknown_types(prog, expect_completion=expect_completion)


def _add_and_infer_dtypes_overdetermined(kernel, dtype_dict):
    kernel = _add_dtypes_overdetermined(kernel, dtype_dict)

    from loopy.type_inference import infer_unknown_types
    return infer_unknown_types(kernel, expect_completion=True)

# }}}


# {{{ find_all_insn_inames fixed point iteration (deprecated)

def guess_iname_deps_based_on_var_use(kernel, insn, insn_id_to_inames=None):
    # For all variables that insn depends on, find the intersection
    # of iname deps of all writers, and add those to insn's
    # dependencies.

    result = frozenset()

    writer_map = kernel.writer_map()

    for tv_name in (insn.read_dependency_names() & kernel.get_written_variables()):
        tv_implicit_inames = None

        for writer_id in writer_map[tv_name]:
            writer_insn = kernel.id_to_insn[writer_id]
            if insn_id_to_inames is None:
                writer_inames = writer_insn.within_inames
            else:
                writer_inames = insn_id_to_inames[writer_id]

            writer_implicit_inames = (
                    writer_inames
                    - (writer_insn.write_dependency_names() & kernel.all_inames()))
            if tv_implicit_inames is None:
                tv_implicit_inames = writer_implicit_inames
            else:
                tv_implicit_inames = (tv_implicit_inames
                        & writer_implicit_inames)

        if tv_implicit_inames is not None:
            result = result | tv_implicit_inames

    return result - insn.reduction_inames()


def find_all_insn_inames(kernel):
    logger.debug("%s: find_all_insn_inames: start" % kernel.name)

    writer_map = kernel.writer_map()

    insn_id_to_inames = {}
    insn_assignee_inames = {}

    all_read_deps = {}
    all_write_deps = {}

    from loopy.transform.subst import expand_subst
    kernel = expand_subst(kernel)

    for insn in kernel.instructions:
        all_read_deps[insn.id] = read_deps = insn.read_dependency_names()
        all_write_deps[insn.id] = write_deps = insn.write_dependency_names()
        deps = read_deps | write_deps

        if insn.within_inames_is_final:
            iname_deps = insn.within_inames
        else:
            iname_deps = (
                    deps & kernel.all_inames()
                    | insn.within_inames)

        assert isinstance(read_deps, frozenset), type(insn)
        assert isinstance(write_deps, frozenset), type(insn)
        assert isinstance(iname_deps, frozenset), type(insn)

        logger.debug("%s: find_all_insn_inames: %s (init): %s - "
                "read deps: %s - write deps: %s" % (
                    kernel.name, insn.id, ", ".join(sorted(iname_deps)),
                    ", ".join(sorted(read_deps)), ", ".join(sorted(write_deps)),
                    ))

        insn_id_to_inames[insn.id] = iname_deps
        insn_assignee_inames[insn.id] = write_deps & kernel.all_inames()

    # fixed point iteration until all iname dep sets have converged

    # Why is fixed point iteration necessary here? Consider the following
    # scenario:
    #
    # z = expr(iname)
    # y = expr(z)
    # x = expr(y)
    #
    # x clearly has a dependency on iname, but this is not found until that
    # dependency has propagated all the way up. Doing this recursively is
    # not guaranteed to terminate because of circular dependencies.

    while True:
        did_something = False
        for insn in kernel.instructions:

            if insn.within_inames_is_final:
                continue

            # {{{ dependency-based propagation

            inames_old = insn_id_to_inames[insn.id]
            inames_new = inames_old | guess_iname_deps_based_on_var_use(
                    kernel, insn, insn_id_to_inames)

            insn_id_to_inames[insn.id] = inames_new

            if inames_new != inames_old:
                did_something = True

                warn_with_kernel(kernel, "inferred_iname",
                        "The iname(s) '%s' on instruction '%s' "
                        "was/were automatically added. "
                        "This is deprecated. Please add the iname "
                        "to the instruction "
                        "explicitly, e.g. by adding 'for' loops"
                        % (", ".join(inames_new-inames_old), insn.id))

            # }}}

            # {{{ domain-based propagation

            inames_old = insn_id_to_inames[insn.id]
            inames_new = set(insn_id_to_inames[insn.id])

            for iname in inames_old:
                home_domain = kernel.domains[kernel.get_home_domain_index(iname)]

                for par in home_domain.get_var_names(dim_type.param):
                    # Add all inames occurring in parameters of domains that my
                    # current inames refer to.

                    if par in kernel.all_inames():
                        inames_new.add(intern(par))

                    # If something writes the bounds of a loop in which I'm
                    # sitting, I had better be in the inames that the writer is
                    # in.

                    if par in kernel.temporary_variables:
                        for writer_id in writer_map.get(par, []):
                            inames_new.update(insn_id_to_inames[writer_id])

            if inames_new != inames_old:
                did_something = True
                insn_id_to_inames[insn.id] = frozenset(inames_new)

                warn_with_kernel(kernel, "inferred_iname",
                        "The iname(s) '%s' on instruction '%s' was "
                        "automatically added. "
                        "This is deprecated. Please add the iname "
                        "to the instruction "
                        "explicitly, e.g. by adding 'for' loops"
                        % (", ".join(inames_new-inames_old), insn.id))

            # }}}

        if not did_something:
            break

    logger.debug("%s: find_all_insn_inames: done" % kernel.name)

    for v in insn_id_to_inames.values():
        assert isinstance(v, frozenset)

    return insn_id_to_inames

# }}}


# {{{ set operation cache

def _eliminate_except(set_, except_inames, dts):
    return set_.eliminate_except(except_inames, dts)


def _get_dim_max(set_, idx):
    return set_.dim_max(idx)


def _get_dim_min(set_, idx):
    return set_.dim_min(idx)


class SetOperationCacheManager:
    def __init__(self):
        # mapping: set hash -> [(set, result)]
        self.cache = {}

    def op(self, set_, op, args):
        hashval = hash((set_, op, args))
        bucket = self.cache.setdefault(hashval, [])

        for bkt_set, result in bucket:
            if set_.plain_is_equal(bkt_set):
                return result

        result = op(set_, *args)
        bucket.append((set_, result))
        return result

    def dim_min(self, set_, *args):
        if set_.plain_is_empty():
            raise LoopyError("domain '%s' is empty" % set_)

        return self.op(set_, _get_dim_min, args)

    def dim_max(self, set_, *args):
        if set_.plain_is_empty():
            raise LoopyError("domain '%s' is empty" % set_)

        return self.op(set_, _get_dim_max, args)

    def eliminate_except(self, set_, *args):
        return self.op(set_, _eliminate_except, args)

    def base_index_and_length(self, set_, iname, context=None,
            n_allowed_params_in_length=None):
        """
        :arg n_allowed_params_in_length: Simplifies the 'length'
            argument so that only the first that many params
            (in the domain of *set_*) occur.
        """
        if not isinstance(iname, int):
            iname_to_dim = set_.space.get_var_dict()
            idx = iname_to_dim[iname][1]
        else:
            idx = iname

        lower_bound_pw_aff = self.dim_min(set_, idx)
        upper_bound_pw_aff = self.dim_max(set_, idx)

        from loopy.diagnostic import StaticValueFindingError
        from loopy.isl_helpers import (
            find_max_of_pwaff_with_params,
            static_max_of_pw_aff,
            static_min_of_pw_aff,
            static_value_of_pw_aff,
        )
        from loopy.symbolic import pw_aff_to_expr

        # {{{ first: try to find static lower bound value

        try:
            base_index_aff = static_value_of_pw_aff(
                    lower_bound_pw_aff, constants_only=False,
                    context=context)
        except StaticValueFindingError:
            base_index_aff = None

        if base_index_aff is not None:
            base_index = pw_aff_to_expr(base_index_aff)

            length = find_max_of_pwaff_with_params(
                    upper_bound_pw_aff - base_index_aff + 1,
                    n_allowed_params_in_length)
            length = pw_aff_to_expr(static_max_of_pw_aff(
                    length, constants_only=False,
                    context=context))

            return base_index, length

        # }}}

        # {{{ if that didn't work, try finding a lower bound

        base_index_aff = static_min_of_pw_aff(
                lower_bound_pw_aff, constants_only=False,
                context=context)

        base_index = pw_aff_to_expr(base_index_aff)

        length = find_max_of_pwaff_with_params(
                upper_bound_pw_aff - base_index_aff + 1,
                n_allowed_params_in_length)
        length = pw_aff_to_expr(static_max_of_pw_aff(
                length, constants_only=False,
                context=context))

        return base_index, length

        # }}}

# }}}


# {{{ domain change helper

class DomainChanger:
    """Helps change the domain responsible for *inames* within a kernel.

    .. note: Does not perform an in-place change!
    """

    def __init__(self, kernel, inames):
        """
        :arg inames: a non-mutable iterable
        """

        self.kernel = kernel
        if inames:
            ldi = kernel.get_leaf_domain_indices(inames)
            if len(ldi) > 1:
                raise RuntimeError("Inames '%s' require more than one leaf "
                        "domain, which makes the domain change that is part "
                        "of your current operation ambiguous." % ", ".join(inames))

            self.leaf_domain_index, = ldi
            self.domain = kernel.domains[self.leaf_domain_index]

        else:
            self.domain = kernel.combine_domains(())
            self.leaf_domain_index = None

    def get_original_domain(self):
        return self.kernel.domains[self.leaf_domain_index]

    def get_domains_with(self, replacement):
        result = self.kernel.domains[:]
        if self.leaf_domain_index is not None:
            result[self.leaf_domain_index] = replacement
        else:
            result.append(replacement)

        return result

    def get_kernel_with(self, replacement):
        return self.kernel.copy(
                domains=self.get_domains_with(replacement),

                # Changing the domain might look like it wants to change grid
                # sizes. Not true.
                # (Relevant for 'slab decomposition')
                overridden_get_grid_sizes_for_insn_ids=(
                    self.kernel.get_grid_sizes_for_insn_ids))

# }}}


# {{{ graphviz / dot export

@for_each_kernel
def get_dot_dependency_graph(kernel, callables_table, iname_cluster=True,
        use_insn_id=False):
    """Return a string in the `dot <https://graphviz.org/>`_ language depicting
    dependencies among kernel instructions.
    """

    # make sure all automatically added stuff shows up
    from loopy.kernel.creation import apply_single_writer_dependency_heuristic
    kernel = apply_single_writer_dependency_heuristic(kernel, warn_if_used=False)

    if iname_cluster and not kernel.linearization:
        try:
            from loopy.schedule import get_one_linearized_kernel
            kernel = get_one_linearized_kernel(kernel, callables_table)
        except RuntimeError as e:
            iname_cluster = False
            from warnings import warn
            warn("error encountered during scheduling for dep graph -- "
                    "cannot perform iname clustering: %s(%s)"
                    % (type(e).__name__, e), stacklevel=1)

    dep_graph = {}
    lines = []

    from loopy.kernel.data import CInstruction, MultiAssignmentBase

    for insn in kernel.instructions:
        if isinstance(insn, MultiAssignmentBase):
            lhs = ", ".join(str(assignee) for assignee in insn.assignees)
            op = f"{lhs} <- {insn.expression}"
            if len(op) > 200:
                op = op[:200] + "..."

        elif isinstance(insn, CInstruction):
            op = "<C instruction %s>" % insn.id
        else:
            op = "<instruction %s>" % insn.id

        if use_insn_id:
            insn_label = insn.id
            tooltip = op
        else:
            insn_label = op
            tooltip = insn.id

        lines.append('"%s" [label="%s",shape="box",tooltip="%s"];'
                % (
                    insn.id,
                    repr(insn_label)[1:-1],
                    repr(tooltip)[1:-1],
                    ))
        for dep in insn.depends_on:
            dep_graph.setdefault(insn.id, set()).add(dep)

    # {{{ O(n^3) transitive reduction

    # first, compute transitive closure by fixed point iteration
    while True:
        changed_something = False

        for insn_1 in dep_graph:
            for insn_2 in dep_graph.get(insn_1, set()).copy():
                for insn_3 in dep_graph.get(insn_2, set()).copy():
                    if insn_3 not in dep_graph.get(insn_1, set()):
                        changed_something = True
                        dep_graph[insn_1].add(insn_3)

        if not changed_something:
            break

    for insn_1 in dep_graph:
        for insn_2 in dep_graph.get(insn_1, set()).copy():
            for insn_3 in dep_graph.get(insn_2, set()).copy():
                if insn_3 in dep_graph.get(insn_1, set()):
                    dep_graph[insn_1].remove(insn_3)

    # }}}

    for insn_1 in dep_graph:
        for insn_2 in dep_graph.get(insn_1, set()):
            lines.append(f"{insn_2} -> {insn_1}")

    if iname_cluster:
        from loopy.schedule import (
            Barrier,
            CallKernel,
            EnterLoop,
            LeaveLoop,
            ReturnFromKernel,
            RunInstruction,
        )

        for sched_item in kernel.linearization:
            if isinstance(sched_item, EnterLoop):
                lines.append('subgraph cluster_%s { label="%s"'
                        % (sched_item.iname, sched_item.iname))
            elif isinstance(sched_item, LeaveLoop):
                lines.append("}")
            elif isinstance(sched_item, RunInstruction):
                lines.append(sched_item.insn_id)
            elif isinstance(sched_item, (CallKernel, ReturnFromKernel, Barrier)):
                pass
            else:
                raise LoopyError("schedule item not unterstood: %r" % sched_item)

    return "digraph {} {{\n{}\n}}".format(
            kernel.name,
            "\n".join(lines)
            )


def show_dependency_graph(*args, **kwargs):
    """Show the dependency graph generated by :func:`get_dot_dependency_graph`
    in a browser. Accepts the same arguments as that function.
    """

    dot = get_dot_dependency_graph(*args, **kwargs)

    from tempfile import mkdtemp
    temp_dir = mkdtemp(prefix="tmp_loopy_dot")

    dot_file_name = "loopy.dot"

    from os.path import join
    with open(join(temp_dir, dot_file_name), "w") as dotf:
        dotf.write(dot)

    svg_file_name = "loopy.svg"
    from subprocess import check_call
    check_call(["dot", "-Tsvg", "-o", svg_file_name, dot_file_name],
            cwd=temp_dir)

    full_svg_file_name = join(temp_dir, svg_file_name)
    logger.info("show_dot_dependency_graph: svg written to '%s'"
            % full_svg_file_name)

    from webbrowser import open as browser_open
    browser_open("file://" + full_svg_file_name)

# }}}


# {{{ is domain dependent on inames

def is_domain_dependent_on_inames(kernel: LoopKernel,
        domain_index: int, inames: AbstractSet[str]) -> bool:
    dom = kernel.domains[domain_index]
    dom_parameters = set(dom.get_var_names(dim_type.param))

    # {{{ check for parenthood by loop bound iname

    if inames & dom_parameters:
        return True

    # }}}

    # {{{ check for parenthood by written variable

    for par in dom_parameters:
        if par in kernel.temporary_variables:
            writer_insns = kernel.writer_map()[par]

            if len(writer_insns) > 1:
                raise RuntimeError("loop bound '%s' "
                        "may only be written to once" % par)

            writer_insn, = writer_insns
            writer_inames = kernel.insn_inames(writer_insn)

            if writer_inames & inames:
                return True

    # }}}

    return False

# }}}


# {{{ assign automatic axes

# {{{ rank inames by stride

def get_auto_axis_iname_ranking_by_stride(kernel, insn):
    from loopy.kernel.data import ImageArg, ValueArg

    approximate_arg_values = {}
    for arg in kernel.args:
        if isinstance(arg, ValueArg):
            if arg.approximately is not None:
                approximate_arg_values[arg.name] = arg.approximately
            else:
                raise LoopyError("No approximate arg value specified for '%s'"
                        % arg.name)

    # {{{ find all array accesses in insn

    from loopy.symbolic import ArrayAccessFinder
    ary_acc_exprs = list(ArrayAccessFinder()(insn.expression))

    from pymbolic.primitives import Subscript

    for assignee in insn.assignees:
        if isinstance(assignee, Subscript):
            ary_acc_exprs.append(assignee)

    # }}}

    # {{{ filter array accesses to only the global ones

    global_ary_acc_exprs = []

    for aae in ary_acc_exprs:
        ary_name = aae.aggregate.name
        arg = kernel.arg_dict.get(ary_name)
        if arg is None:
            continue

        if isinstance(arg, ImageArg):
            continue

        global_ary_acc_exprs.append(aae)

    # }}}

    # {{{ figure out automatic-axis inames

    from loopy.kernel.data import AutoLocalInameTagBase
    auto_axis_inames = {
        iname for iname in insn.within_inames
        if kernel.iname_tags_of_type(iname, AutoLocalInameTagBase)}

    # }}}

    # {{{ figure out which iname should get mapped to local axis 0

    # maps inames to "aggregate stride"
    aggregate_strides = {}

    from pymbolic.primitives import Variable

    from loopy.symbolic import CoefficientCollector

    for aae in global_ary_acc_exprs:
        index_expr = aae.index
        if not isinstance(index_expr, tuple):
            index_expr = (index_expr,)

        ary_name = aae.aggregate.name
        arg = kernel.arg_dict.get(ary_name)

        if arg.dim_tags is None:
            from warnings import warn
            warn("Strides for '%s' are not known. Local axis assignment "
                    "is likely suboptimal." % arg.name, stacklevel=1)
            ary_strides = [1] * len(index_expr)
        else:
            ary_strides = []
            from loopy.kernel.array import FixedStrideArrayDimTag
            for dim_tag in arg.dim_tags:
                if isinstance(dim_tag, FixedStrideArrayDimTag):
                    ary_strides.append(dim_tag.stride)

        # {{{ construct iname_to_stride_expr

        iname_to_stride_expr = {}
        for iexpr_i, stride in zip(index_expr, ary_strides):
            if stride is None:
                continue
            coeffs = CoefficientCollector()(iexpr_i)
            for var, coeff in coeffs.items():
                # This is a nested if instead of 'and' for pylint's benefit.
                if isinstance(var, Variable):
                    if var.name in auto_axis_inames:
                        # excludes '1', i.e.  the constant
                        new_stride = coeff*stride
                        old_stride = iname_to_stride_expr.get(var.name, None)
                        if old_stride is None or new_stride < old_stride:
                            iname_to_stride_expr[var.name] = new_stride

        # }}}

        from pymbolic import evaluate
        for iname, stride_expr in iname_to_stride_expr.items():
            stride = evaluate(stride_expr, approximate_arg_values)
            aggregate_strides[iname] = aggregate_strides.get(iname, 0) + stride

    if aggregate_strides:
        very_large_stride = int(np.iinfo(np.int32).max)

        return sorted((iname for iname in insn.within_inames),
                key=lambda iname: (
                    aggregate_strides.get(iname, very_large_stride),
                    iname))
    else:
        return None

    # }}}

# }}}


def assign_automatic_axes(kernel, callables_table, axis=0, local_size=None):
    logger.debug("%s: assign automatic axes" % kernel.name)
    # TODO: do the tag removal rigorously, might be easier after switching
    # to set() from tuple()

    from loopy.kernel.data import (
        AutoLocalInameTagBase,
        LocalInameTag,
        filter_iname_tags_by_type,
    )

    # Realize that at this point in time, axis lengths are already
    # fixed. So we compute them once and pass them to our recursive
    # copies.

    if local_size is None:
        _, local_size = kernel.get_grid_size_upper_bounds_as_exprs(
                callables_table, ignore_auto=True)

    # {{{ axis assignment helper function

    def assign_axis(recursion_axis, iname, axis=None):
        """Assign iname to local axis *axis* and start over by calling
        the surrounding function assign_automatic_axes.

        If *axis* is None, find a suitable axis automatically.
        """
        try:
            desired_length = kernel.get_constant_iname_length(iname)
        except isl.Error:
            # Likely unbounded, automatic assignment is not
            # going to happen for this iname.
            new_inames = kernel.inames.copy()
            new_inames[iname] = kernel.inames[iname].copy(
                    tags=frozenset(tag
                        for tag in kernel.inames[iname].tags
                        if not isinstance(tag, AutoLocalInameTagBase)))
            return assign_automatic_axes(
                    kernel.copy(inames=new_inames),
                    callables_table,
                    axis=recursion_axis)

        if axis is None:
            # {{{ find a suitable axis

            shorter_possible_axes = []
            test_axis = 0
            while True:
                if test_axis >= len(local_size):
                    break
                if test_axis in assigned_local_axes:
                    test_axis += 1
                    continue

                if local_size[test_axis] < desired_length:
                    shorter_possible_axes.append(test_axis)
                    test_axis += 1
                    continue
                else:
                    axis = test_axis
                    break

            # The loop above will find an unassigned local axis
            # that has enough 'room' for the iname. In the same traversal,
            # it also finds theoretically assignable axes that are shorter,
            # in the variable shorter_possible_axes.

            if axis is None and shorter_possible_axes:
                # sort as longest first
                shorter_possible_axes.sort(key=lambda ax: local_size[ax])
                axis = shorter_possible_axes[0]

            # }}}

        if axis is None:
            new_tag = None
        else:
            new_tag = LocalInameTag(axis)
            if desired_length > local_size[axis]:
                from loopy import untag_inames
                from loopy.transform.iname import split_iname

                # Don't be tempted to switch the outer tag to unroll--this may
                # generate tons of code on some examples.

                return assign_automatic_axes(
                        split_iname(
                            untag_inames(kernel, iname, AutoLocalInameTagBase),
                            iname, inner_length=local_size[axis],
                            outer_tag=None, inner_tag=new_tag,
                            do_tagged_check=False),
                        callables_table=callables_table,
                        axis=recursion_axis, local_size=local_size)

        if not kernel.iname_tags_of_type(iname, AutoLocalInameTagBase):
            raise LoopyError("trying to reassign '%s'" % iname)

        if new_tag:
            new_tag_set = frozenset([new_tag])
        else:
            new_tag_set = frozenset()
        new_tags = (
                frozenset(tag for tag in kernel.inames[iname].tags
                    if not isinstance(tag, AutoLocalInameTagBase))
                | new_tag_set)
        new_inames = kernel.inames.copy()
        new_inames[iname] = kernel.inames[iname].copy(tags=new_tags)
        return assign_automatic_axes(kernel.copy(inames=new_inames),
                callables_table, axis=recursion_axis, local_size=local_size)

    # }}}

    # {{{ main assignment loop

    # assignment proceeds in one phase per axis, each time assigning the
    # smallest-stride available iname to the current axis

    import loopy as lp

    for insn in kernel.instructions:
        if not isinstance(insn, lp.MultiAssignmentBase):
            continue

        auto_axis_inames = [
            iname for iname in insn.within_inames
            if kernel.iname_tags_of_type(iname, AutoLocalInameTagBase)]

        if not auto_axis_inames:
            continue

        assigned_local_axes = set()

        for iname in insn.within_inames:
            tags = kernel.iname_tags_of_type(iname, LocalInameTag, max_num=1)
            if tags:
                tag, = tags
                assigned_local_axes.add(tag.axis)

        if axis < len(local_size):
            # "valid" pass: try to assign a given axis

            if axis not in assigned_local_axes:
                iname_ranking = get_auto_axis_iname_ranking_by_stride(kernel, insn)
                if iname_ranking is not None:
                    for iname in iname_ranking:
                        prev_tags = kernel.iname_tags(iname)
                        if filter_iname_tags_by_type(
                                prev_tags, AutoLocalInameTagBase):
                            return assign_axis(axis, iname, axis)

        else:
            # "invalid" pass: There are still unassigned axis after the
            #  numbered "valid" passes--assign the remainder by length.

            def get_iname_length(iname):
                try:
                    return kernel.get_constant_iname_length(iname)
                except isl.Error:
                    return -1
            # assign longest auto axis inames first
            auto_axis_inames.sort(
                            key=get_iname_length,
                            reverse=True)

            if auto_axis_inames:
                return assign_axis(axis, auto_axis_inames.pop())

    # }}}

    # We've seen all instructions and not punted to recursion/restart because
    # of a new axis assignment.

    if axis >= len(local_size):
        return kernel
    else:
        return assign_automatic_axes(kernel,
                callables_table=callables_table, axis=axis+1,
                local_size=local_size)

# }}}


# {{{ array modifier

class ArrayChanger:
    def __init__(self, kernel, array_name):
        self.kernel = kernel
        self.array_name = array_name

    def get(self):
        ary_name = self.array_name
        if ary_name in self.kernel.temporary_variables:
            result = self.kernel.temporary_variables[ary_name]
        elif ary_name in self.kernel.arg_dict:
            result = self.kernel.arg_dict[ary_name]
        else:
            raise NameError("array '%s' was not found" % ary_name)

        from loopy.kernel.array import ArrayBase
        if not isinstance(result, ArrayBase):
            raise LoopyError("variable '%s' is not an array" % ary_name)

        return result

    def with_changed_array(self, new_array):
        knl = self.kernel
        ary_name = self.array_name

        if ary_name in knl.temporary_variables:
            new_tv = knl.temporary_variables.copy()
            new_tv[ary_name] = new_array
            return knl.copy(temporary_variables=new_tv)

        elif ary_name in knl.arg_dict:
            new_args = []
            for arg in knl.args:
                if arg.name == ary_name:
                    new_args.append(new_array)
                else:
                    new_args.append(arg)

            return knl.copy(args=new_args)

        else:
            raise NameError("array '%s' was not found" % ary_name)

# }}}


# {{{ guess_var_shape

def guess_var_shape(kernel, var_names):
    from loopy.symbolic import BatchedAccessMapMapper, SubstitutionRuleExpander

    armap = BatchedAccessMapMapper(kernel, var_names)

    submap = SubstitutionRuleExpander(kernel.substitutions)

    def run_through_armap(expr):
        armap(submap(expr), insn.within_inames)
        return expr

    try:
        for insn in kernel.instructions:
            insn.with_transformed_expressions(run_through_armap)
    except TypeError as e:
        from traceback import print_exc
        print_exc()

        raise LoopyError(
                "Failed to (automatically, as requested) find "
                "shape/strides for variables '%s'. "
                "Specifying the shape manually should get rid of this. "
                "The following error occurred: %s"
                % (",".join(var_names), str(e))) from e

    result = []
    for var_name in var_names:
        access_range = armap.get_access_range(var_name)
        bad_subscripts = armap.bad_subscripts[var_name]
        if access_range is None:
            if bad_subscripts:
                from loopy.symbolic import LinearSubscript
                if any(isinstance(sub, LinearSubscript)
                        for sub in bad_subscripts):
                    raise LoopyError("cannot determine access range for '%s': "
                            "linear subscript(s) in '%s'"
                            % (var_name, ", ".join(
                                    str(i) for i in bad_subscripts)))

                n_axes_in_subscripts = {
                        len(sub.index_tuple) for sub in bad_subscripts}

                if len(n_axes_in_subscripts) != 1:
                    raise RuntimeError("subscripts of '%s' with differing "
                            "numbers of axes were found" % var_name)

                n_axes, = n_axes_in_subscripts

                if n_axes == 1:
                    # Leave shape undetermined--we can live with that for 1D.
                    shape = None
                else:
                    raise LoopyError("cannot determine access range for '%s': "
                            "undetermined index in subscript(s) '%s'"
                            % (var_name, ", ".join(
                                    str(i) for i in bad_subscripts)))

            else:
                # no subscripts found, let's call it a scalar
                shape = ()
        else:
            from loopy.isl_helpers import static_max_of_pw_aff
            from loopy.symbolic import pw_aff_to_expr

            shape = []
            for i in range(access_range.dim(dim_type.set)):
                try:
                    shape.append(
                            pw_aff_to_expr(static_max_of_pw_aff(
                                kernel.cache_manager.dim_max(
                                    access_range, i) + 1,
                                constants_only=False)))
                except Exception:
                    print("While trying to find shape axis %d of "
                            "variable '%s', the following "
                            "exception occurred:" % (i, var_name),
                            file=sys.stderr)
                    print("*** ADVICE: You may need to manually specify the "
                            "shape of argument '%s'." % (var_name),
                            file=sys.stderr)
                    raise

            shape = tuple(shape)
        result.append(shape)

    return tuple(result)

# }}}


# {{{ loop nest tracker

class SetTrie:
    """
    Similar to a trie, but uses an unordered sequence as the key.
    """

    def __init__(self, children=(), all_items=None):
        self.children = dict(children)
        # all_items should be shared within a trie.
        if all_items is None:
            self.all_items = set()
        else:
            self.all_items = all_items

    def descend(self, on_found=lambda prefix: None, prefix=frozenset()):
        on_found(prefix)
        for prefix, child in sorted(
                self.children.items(),
                key=lambda it: sorted(it[0])):
            child.descend(on_found, prefix=prefix)

    def check_consistent_insert(self, items_to_insert):
        if items_to_insert & self.all_items:
            raise ValueError("inconsistent nesting")

    def add_or_update(self, key):
        if len(key) == 0:
            return

        for child_key, child in self.children.items():  # noqa: B007
            common = child_key & key
            if common:
                break
        else:
            # Key not found - insert new child
            self.check_consistent_insert(key)
            self.children[frozenset(key)] = SetTrie(all_items=self.all_items)
            self.all_items.update(key)
            return

        if child_key <= key:
            # child is a prefix of key:
            child.add_or_update(key - common)
        elif key < child_key:
            # key is a strict prefix of child:
            #
            #  -[new child]
            #     |
            #   [child]
            #
            del self.children[child_key]
            self.children[common] = SetTrie(
                children={frozenset(child_key - common): child},
                all_items=self.all_items)
        else:
            # key and child share a common prefix:
            #
            # -[new placeholder]
            #      /        \
            #  [new child]   [child]
            #
            self.check_consistent_insert(key - common)

            del self.children[child_key]
            self.children[common] = SetTrie(
                children={
                    frozenset(child_key - common): child,
                    frozenset(key - common): SetTrie(all_items=self.all_items)},
                all_items=self.all_items)
            self.all_items.update(key - common)


def get_visual_iname_order_embedding(kernel):
    """
    Return :class:`dict` `embedding` mapping inames to a totally ordered set of
    values, such that `embedding[iname1] < embedding[iname2]` when `iname2`
    is nested inside `iname1`.
    """
    from loopy.kernel.data import IlpBaseTag
    # Ignore ILP tagged inames, since they do not have to form a strict loop
    # nest.
    ilp_inames = frozenset(iname
        for iname in kernel.inames
        if kernel.iname_tags_of_type(iname, IlpBaseTag))

    iname_trie = SetTrie()

    for insn in kernel.instructions:
        within_inames = {
            iname for iname in insn.within_inames
            if iname not in ilp_inames}
        iname_trie.add_or_update(within_inames)

    embedding = {}

    def update_embedding(inames):
        embedding.update(
            {iname: (len(embedding), iname) for iname in inames})

    iname_trie.descend(update_embedding)

    for iname in ilp_inames:
        # Nest ilp_inames innermost, so they don't interrupt visual order.
        embedding[iname] = (len(embedding), iname)

    return embedding

# }}}


# {{{ find_recursive_dependencies

def find_recursive_dependencies(kernel, insn_ids):
    queue = list(insn_ids)

    result = set(insn_ids)

    while queue:
        new_queue = []

        for insn_id in queue:
            insn = kernel.id_to_insn[insn_id]
            additional = insn.depends_on - result
            result.update(additional)
            new_queue.extend(additional)

        queue = new_queue

    return result

# }}}


# {{{ find_reverse_dependencies

def find_reverse_dependencies(kernel, insn_ids):
    """Finds a set of IDs of instructions that depend on one of the insn_ids.

    :arg insn_ids: a set of instruction IDs
    """
    return frozenset(
            insn.id
            for insn in kernel.instructions
            if insn.depends_on & insn_ids)

# }}}


# {{{ draw_dependencies_as_unicode_arrows

def draw_dependencies_as_unicode_arrows(
        instructions, fore, style, flag_downward=True, max_columns=20):
    """
    :arg instructions: an ordered iterable of :class:`loopy.InstructionBase`
        instances
    :arg fore: if given, will be used like a :mod:`colorama` ``Fore`` object
        to color-code dependencies. (E.g. red for downward edges)
    :returns: A tuple ``(uniform_length, rows)``, where *rows* is a list of
        tuples (arrows, extender) with Unicode-drawn dependency arrows, one per
        entry of *instructions*. *extender* can be used to extend arrows below the
        line of an instruction. *uniform_length* is the length of the *arrows* and
        *extender* strings.
    """
    reverse_deps = {}

    for insn in instructions:
        for dep in insn.depends_on:
            reverse_deps.setdefault(dep, set()).add(insn.id)

    # mapping of to_id tuples to column_index
    dep_to_column = {}

    # {{{ find column assignments

    # mapping from column indices to (end_insn_ids, pointed_at_insn_id)
    # end_insn_ids is a set that gets modified in-place to remove 'ends'
    # (arrow origins) as they are being passed.
    columns_in_use = {}

    n_columns = [0]

    def find_free_column():
        i = 0
        while i in columns_in_use:
            i += 1
        if i+1 > n_columns[0]:
            n_columns[0] = i+1
            row.append(" ")
        return i

    def do_flag_downward(s, pointed_at_insn_id):
        if flag_downward and pointed_at_insn_id not in processed_ids:
            return fore.RED+s+style.RESET_ALL
        else:
            return s

    def make_extender():
        result = n_columns[0] * [" "]
        for col, (_, pointed_at_insn_id) in columns_in_use.items():
            result[col] = do_flag_downward("│", pointed_at_insn_id)

        return result

    processed_ids = set()

    rows = []
    for insn in instructions:
        row = make_extender()

        # {{{ add rdeps for already existing columns

        rdeps = reverse_deps.get(insn.id, set()).copy() - processed_ids
        assert insn.id not in rdeps

        col = dep_to_column.get(insn.id)
        if col is not None:
            columns_in_use[col][0].update(rdeps)
        del col

        # }}}

        # {{{ add deps for already existing columns

        for dep in insn.depends_on:
            dep_key = dep
            if dep_key in dep_to_column:
                col = dep_to_column[dep]
                columns_in_use[col][0].add(insn.id)

        # }}}

        for col, (starts, pointed_at_insn_id) in list(columns_in_use.items()):
            if insn.id == pointed_at_insn_id:
                if starts:
                    # will continue downward
                    row[col] = do_flag_downward(">", pointed_at_insn_id)
                else:
                    # stops here

                    # placeholder, pending deletion
                    columns_in_use[col] = None

                    row[col] = do_flag_downward("↳", pointed_at_insn_id)

            elif insn.id in starts:
                starts.remove(insn.id)
                if starts or pointed_at_insn_id not in processed_ids:
                    # will continue downward
                    row[col] = do_flag_downward("├", pointed_at_insn_id)

                else:
                    # stops here
                    row[col] = "└"
                    # placeholder, pending deletion
                    columns_in_use[col] = None

        # {{{ start arrows by reverse dep

        dep_key = insn.id
        if dep_key not in dep_to_column and rdeps:
            col = dep_to_column[dep_key] = find_free_column()
            columns_in_use[col] = (rdeps, insn.id)
            row[col] = "↱"

        # }}}

        # {{{ start arrows by forward dep

        for dep in insn.depends_on:
            assert dep != insn.id
            dep_key = dep
            if dep_key not in dep_to_column:
                col = dep_to_column[dep_key] = find_free_column()

                # No need to add current instruction to end_insn_ids set, as
                # we're currently handling it.
                columns_in_use[col] = (set(), dep)

                row[col] = do_flag_downward("┌", dep)

        # }}}

        # {{{ delete columns_in_use entry for end-of-life columns

        for col, value in list(columns_in_use.items()):
            if value is None:
                del columns_in_use[col]

        # }}}

        processed_ids.add(insn.id)

        extender = make_extender()

        rows.append(("".join(row), "".join(extender)))

    # }}}

    uniform_length = min(n_columns[0], max_columns)

    added_ellipsis = [False]

    def len_without_color_escapes(s):
        s = (s
                .replace(fore.RED, "")
                .replace(style.RESET_ALL, ""))
        return len(s)

    def truncate_without_color_escapes(s, length):
        # FIXME: This is a bit dumb--it removes color escapes when truncation
        # is needed.

        s = (s
                .replace(fore.RED, "")
                .replace(style.RESET_ALL, ""))

        return s[:length] + "…"

    def conform_to_uniform_length(s):
        len_s = len_without_color_escapes(s)

        if len_s <= uniform_length:
            return s + " "*(uniform_length-len_s)
        else:
            added_ellipsis[0] = True
            return truncate_without_color_escapes(s, uniform_length)

    rows = [
            (conform_to_uniform_length(row),
                conform_to_uniform_length(extender))
            for row, extender in rows]

    if added_ellipsis[0]:
        uniform_length += 1

        rows = [
                (conform_to_uniform_length(row),
                    conform_to_uniform_length(extender))
                for row, extender in rows]

    return uniform_length, rows

# }}}


# {{{ stringify_instruction_list

def stringify_instruction_tag(tag: Tag) -> str:
    from loopy.kernel.instruction import LegacyStringInstructionTag
    if isinstance(tag, LegacyStringInstructionTag):
        return f"S({tag.value})"
    else:
        return str(tag)


def stringify_instruction_list(kernel: LoopKernel) -> list[str]:
    # {{{ topological sort

    printed_insn_ids = set()
    printed_insn_order = []

    def insert_insn_into_order(insn):
        if insn.id in printed_insn_ids:
            # Note: dependency cycles are deliberately ignored so that printing
            # succeeds.
            return
        printed_insn_ids.add(insn.id)

        for dep_id in natsorted(insn.depends_on):
            insert_insn_into_order(kernel.id_to_insn[dep_id])

        printed_insn_order.append(insn)

    for insn in kernel.instructions:
        insert_insn_into_order(insn)

    # }}}

    import loopy as lp

    Fore = kernel.options._fore  # noqa
    Style = kernel.options._style  # noqa

    uniform_arrow_length, arrows_and_extenders = \
            draw_dependencies_as_unicode_arrows(
                    printed_insn_order, fore=Fore, style=Style)

    leader = " " * uniform_arrow_length
    lines = []
    current_inames: list[set[str]] = [set()]

    if uniform_arrow_length:
        indent_level = [1]
    else:
        indent_level = [0]

    indent_increment = 2

    iname_order = kernel._get_iname_order_for_printing()

    def add_pre_line(s: str) -> None:
        lines.append(leader + " " * indent_level[0] + s)

    def add_main_line(s: str) -> None:
        lines.append(arrows + " " * indent_level[0] + s)

    def add_post_line(s: str) -> None:
        lines.append(extender + " " * indent_level[0] + s)

    def adapt_to_new_inames_list(new_inames):
        added = []
        removed = []

        # FIXME: Doesn't respect strict nesting
        for iname in iname_order:
            is_in_current = iname in current_inames[0]
            is_in_new = iname in new_inames

            if is_in_new == is_in_current:
                pass
            elif is_in_new and not is_in_current:
                added.append(iname)
            elif not is_in_new and is_in_current:
                removed.append(iname)
            else:
                raise AssertionError()

        if removed:
            indent_level[0] -= indent_increment * len(removed)
            add_pre_line("end " + ", ".join(removed))
        if added:
            add_pre_line("for " + ", ".join(added))
            indent_level[0] += indent_increment * len(added)

        current_inames[0] = new_inames

    for insn, (arrows, extender) in zip(  # noqa: B007
            printed_insn_order, arrows_and_extenders):
        if isinstance(insn, lp.MultiAssignmentBase):
            lhs = ", ".join(str(a) for a in insn.assignees)
            rhs = str(insn.expression)
            trailing = []
        elif isinstance(insn, lp.CInstruction):
            lhs = ", ".join(str(a) for a in insn.assignees)
            rhs = "CODE({}|{})".format(
                    ", ".join(str(x) for x in insn.read_variables),
                    ", ".join(f"{name}={expr}"
                        for name, expr in insn.iname_exprs))

            trailing = insn.code.split("\n")
        elif isinstance(insn, lp.BarrierInstruction):
            lhs = ""
            rhs = "... %sbarrier" % insn.synchronization_kind[0]
            trailing = []

        elif isinstance(insn, lp.NoOpInstruction):
            lhs = ""
            rhs = "... nop"
            trailing = []

        else:
            raise LoopyError("unexpected instruction type: %s"
                    % type(insn).__name__)

        adapt_to_new_inames_list(insn.within_inames)

        options = ["id="+Fore.GREEN+insn.id+Style.RESET_ALL]
        if insn.priority:
            options.append("priority=%d" % insn.priority)
        if insn.tags:
            options.append("tags=%s" % ":".join(
                stringify_instruction_tag(t) for t in insn.tags))
        if isinstance(insn, lp.Assignment) and insn.atomicity:
            options.append("atomic=%s" % ":".join(
                str(a) for a in insn.atomicity))
        if insn.groups:
            options.append("groups=%s" % ":".join(insn.groups))
        if insn.conflicts_with_groups:
            options.append(
                    "conflicts=%s" % ":".join(insn.conflicts_with_groups))
        if insn.no_sync_with:
            options.append("no_sync_with=%s" % ":".join(
                "%s@%s" % entry for entry in sorted(insn.no_sync_with)))
        if isinstance(insn, lp.BarrierInstruction) and \
                insn.synchronization_kind == "local":
            options.append("mem_kind=%s" % insn.mem_kind)

        if lhs:
            core = "{} = {}".format(
                Fore.CYAN+lhs+Style.RESET_ALL,
                Fore.MAGENTA+rhs+Style.RESET_ALL,
                )
        else:
            core = Fore.MAGENTA+rhs+Style.RESET_ALL

        options_str = "  {%s}" % ", ".join(options)

        if insn.predicates:
            # FIXME: precedence
            add_pre_line("if %s" % " and ".join([str(x) for x in insn.predicates]))
            indent_level[0] += indent_increment

        add_main_line(core + options_str)

        for t in trailing:
            add_post_line(t)

        if insn.predicates:
            indent_level[0] -= indent_increment
            add_post_line("end")

        leader = extender

    adapt_to_new_inames_list([])

    return lines

# }}}


# {{{ global barrier order finding

def _insn_id_is_global_barrier(insn_id, kernel):
    insn = kernel.id_to_insn[insn_id]
    return _insn_is_global_barrier(insn)


def _insn_is_global_barrier(insn):
    from loopy.kernel.instruction import BarrierInstruction
    return isinstance(insn, BarrierInstruction) and \
        insn.synchronization_kind == "global"


@memoize_on_first_arg
def kernel_has_global_barriers(kernel):
    return any(_insn_is_global_barrier(insn) for insn in kernel.instructions)


@memoize_on_first_arg
def get_global_barrier_order(kernel):
    """Return a :class:`tuple` of the listing the ids of global barrier instructions
    as they appear in order in the kernel.

    See also :class:`loopy.instruction.BarrierInstruction`.
    """
    dep_graph = {insn.id: set() for insn in kernel.instructions}
    for insn in kernel.instructions:
        for dep in insn.depends_on:
            dep_graph[dep].add(insn.id)

    from pytools.graph import compute_topological_order
    order = compute_topological_order(dep_graph)

    barriers = [
            insn_id for insn_id in order
            if _insn_id_is_global_barrier(insn_id, kernel)]

    del order

    # Ensure this is the only possible order.
    #
    # We do this by looking at the barriers in order.
    # We check for each adjacent pair (a,b) in the order if a < b,
    # i.e. if a is reachable by a chain of dependencies from b.
    visited = set()
    visiting = set()

    for prev_barrier, barrier in itertools.pairwise(barriers):
        # Check if prev_barrier is reachable from barrier.
        stack = [barrier]
        visited.discard(prev_barrier)

        while stack:
            top = stack[-1]

            if top in visiting:
                visiting.remove(top)

            if top in visited:
                stack.pop()
                continue

            visited.add(top)
            visiting.add(top)

            if top == prev_barrier:
                visiting.clear()
                break

            for child in kernel.id_to_insn[top].depends_on:
                stack.append(child)
        else:
            # Search exhausted and we did not find prev_barrier.
            raise LoopyError("barriers '%s' and '%s' are not ordered"
                             % (prev_barrier, barrier))

    return tuple(barriers)

# }}}


# {{{ find most recent global barrier

@memoize_on_first_arg
def find_most_recent_global_barrier(kernel, insn_id):
    """Return the id of the latest occurring global barrier which the
    given instruction (indirectly or directly) depends on, or *None* if this
    instruction does not depend on a global barrier.

    The return value is guaranteed to be unique because global barriers are
    totally ordered within the kernel.
    """

    insn = kernel.id_to_insn[insn_id]

    if len(insn.depends_on) == 0:
        return None

    if not kernel_has_global_barriers(kernel):
        return None

    global_barrier_order = get_global_barrier_order(kernel)

    if len(global_barrier_order) == 0:
        return None

    global_barrier_to_ordinal = {
            b: i for i, b in enumerate(global_barrier_order)}

    def get_barrier_ordinal(barrier_id):
        return (global_barrier_to_ordinal[barrier_id]
                if barrier_id is not None
                else -1)

    direct_barrier_dependencies = {
        dep for dep in insn.depends_on if _insn_id_is_global_barrier(dep, kernel)}

    if len(direct_barrier_dependencies) > 0:
        return max(direct_barrier_dependencies, key=get_barrier_ordinal)
    else:
        return max((find_most_recent_global_barrier(kernel, dep)
                    for dep in insn.depends_on),
                key=get_barrier_ordinal)

# }}}


# {{{ subkernel tools

@memoize_on_first_arg
def get_subkernel_start_indices(kernel: LoopKernel) -> Mapping[str, int]:
    """Return a sequence of schedule indices of

    See also :class:`loopy.schedule.CallKernel`.
    """
    from loopy.kernel import KernelState
    if kernel.state != KernelState.LINEARIZED:
        raise LoopyError("Kernel must be linearized")

    assert kernel.linearization is not None

    from loopy.schedule import CallKernel

    return {lin_item.kernel_name: i
            for i, lin_item in enumerate(kernel.linearization)
            if isinstance(lin_item, CallKernel)}


@memoize_on_first_arg
def get_subkernels(kernel) -> Sequence[str]:
    """Return a :class:`tuple` of the names of the subkernels in the kernel. The
    kernel must be scheduled.

    See also :class:`loopy.schedule.CallKernel`.
    """
    from loopy.kernel import KernelState
    if kernel.state != KernelState.LINEARIZED:
        raise LoopyError("Kernel must be linearized")

    assert kernel.linearization is not None

    return tuple(kernel.linearization[i].kernel_name
            for i in get_subkernel_start_indices(kernel).values())


@memoize_on_first_arg
def get_subkernel_to_insn_id_map(kernel: LoopKernel) -> Mapping[str, frozenset[str]]:
    """Return a :class:`dict` mapping subkernel names to a :class:`frozenset`
    consisting of the instruction ids scheduled within the subkernel. The
    kernel must be scheduled.
    """
    from loopy.kernel import KernelState
    if kernel.state != KernelState.LINEARIZED:
        raise LoopyError("Kernel must be scheduled")

    assert kernel.linearization is not None

    from loopy.schedule import CallKernel, ReturnFromKernel, sched_item_to_insn_id

    subkernel = None
    result: dict[str, set[str]] = {}

    for lin_item in kernel.linearization:
        if isinstance(lin_item, CallKernel):
            subkernel = lin_item.kernel_name
            result[subkernel] = set()

        if isinstance(lin_item, ReturnFromKernel):
            subkernel = None

        if subkernel is not None:
            for insn_id in sched_item_to_insn_id(lin_item):
                result[subkernel].add(insn_id)

    return {name: frozenset(insn_ids) for name, insn_ids in result.items()}


@memoize_on_first_arg
def get_subkernel_extra_inames(kernel: LoopKernel) -> Mapping[str, frozenset[str]]:
    from loopy.kernel import KernelState
    if kernel.state != KernelState.LINEARIZED:
        raise LoopyError("Kernel must be scheduled")

    assert kernel.linearization is not None

    result = {}
    inames: list[str] = []

    from loopy.schedule import CallKernel, EnterLoop, LeaveLoop

    for lin_item in kernel.linearization:
        if isinstance(lin_item, CallKernel):
            result[lin_item.kernel_name] = frozenset(inames)
        elif isinstance(lin_item, EnterLoop):
            inames.append(lin_item.iname)
        elif isinstance(lin_item, LeaveLoop):
            inames.pop()

    return result

# }}}


# {{{ find aliasing equivalence classes

class DisjointSets:
    """
    .. automethod:: __getitem__
    .. automethod:: find_leader_or_create_group
    .. automethod:: union
    .. automethod:: union_many
    """

    # https://en.wikipedia.org/wiki/Disjoint-set_data_structure

    def __init__(self):
        self.leader_to_group = {}
        self.element_to_leader = {}

    def __getitem__(self, item):
        """
        :arg item: A representative of an equivalence class.
        :returns: the equivalence class, given as a set of elements
        """
        try:
            leader = self.element_to_leader[item]
        except KeyError:
            return {item}
        else:
            return self.leader_to_group[leader]

    def find_leader_or_create_group(self, el):
        try:
            return self.element_to_leader[el]
        except KeyError:
            pass

        self.element_to_leader[el] = el
        self.leader_to_group[el] = {el}
        return el

    def union(self, a, b):
        leader_a = self.find_leader_or_create_group(a)
        leader_b = self.find_leader_or_create_group(b)

        if leader_a == leader_b:
            return

        new_leader = leader_a

        for b_el in self.leader_to_group[leader_b]:
            self.element_to_leader[b_el] = new_leader

        self.leader_to_group[leader_a].update(self.leader_to_group[leader_b])
        del self.leader_to_group[leader_b]

    def union_many(self, relation):
        """
        :arg relation: an iterable of 2-tuples enumerating the elements of the
            relation. The relation is assumed to be an equivalence relation
            (transitive, reflexive, symmetric) but need not explicitly contain
            all elements to make it that.

            The first elements of the tuples become group leaders.

        :returns: *self*
        """

        for a, b in relation:
            self.union(a, b)

        return self


def find_aliasing_equivalence_classes(kernel):
    return DisjointSets().union_many(
            (tv.base_storage, tv.name)
            for tv in kernel.temporary_variables.values()
            if tv.base_storage is not None)

# }}}


# {{{ direction helper tools

def infer_args_are_input_output(kernel):
    """
    Returns a copy of *kernel* with the attributes ``is_input`` and
    ``is_output`` of the arguments set.

    .. note::

        If the :attr:`~loopy.ArrayArg.is_output` is not supplied from a user,
        then the array is inferred as an output argument if it is written at
        some point in the kernel.

        If the :attr:`~loopy.ArrayArg.is_input` is not supplied from a user,
        then the array is inferred as an input argument if it is either read at
        some point in the kernel or it is neither read nor written.
    """
    from loopy.kernel.data import ArrayArg, ConstantArg, ImageArg, ValueArg
    new_args = []

    for arg in kernel.args:
        if isinstance(arg, ArrayArg):
            if arg.is_output is not None:
                assert isinstance(arg.is_output, bool)
            else:
                if arg.name in kernel.get_written_variables():
                    arg = arg.copy(is_output=True)
                else:
                    arg = arg.copy(is_output=False)

            if arg.is_input is not None:
                assert isinstance(arg.is_input, bool)
            else:
                if arg.name in kernel.get_read_variables() or (
                        (arg.name not in kernel.get_read_variables()) and (
                            arg.name not in kernel.get_written_variables())):
                    arg = arg.copy(is_input=True)
                else:
                    arg = arg.copy(is_input=False)
        elif isinstance(arg, (ConstantArg, ImageArg, ValueArg)):
            pass
        else:
            raise NotImplementedError("Unknown argument type %s." % type(arg))

        if not (arg.is_input or arg.is_output):
            raise LoopyError("Kernel argument must be either input or output."
                    " '{}' in '{}' does not follow it.".format(arg.name,
                        kernel.name))

        new_args.append(arg)

    return kernel.copy(args=new_args)

# }}}


# {{{ CallablesIDCollector

class CallablesIDCollector(CombineMapper):
    """
    Mapper to collect function identifiers of all resolved callables in an
    expression.
    """
    def combine(self, values):
        import operator
        return reduce(operator.or_, values, frozenset())

    def map_resolved_function(self, expr):
        return frozenset([expr.name])

    def map_constant(self, expr):
        return frozenset()

    def map_kernel(self, kernel):
        callables_in_insn = frozenset()

        for insn in kernel.instructions:
            if isinstance(insn, MultiAssignmentBase):
                callables_in_insn = callables_in_insn | (
                        self(insn.expression))
            elif isinstance(insn, (CInstruction, _DataObliviousInstruction)):
                pass
            else:
                raise NotImplementedError(type(insn).__name__)

        for rule in kernel.substitutions.values():
            callables_in_insn = callables_in_insn | (
                    self(rule.expression))

        return callables_in_insn

    def map_type_cast(self, expr):
        return self.rec(expr.child)

    map_variable = map_constant
    map_function_symbol = map_constant
    map_tagged_variable = map_constant
    map_nan = map_constant


def get_resolved_callable_ids_called_by_knl(knl, callables, recursive=True):
    clbl_id_collector = CallablesIDCollector()
    callables_called_by_kernel = clbl_id_collector.map_kernel(knl)

    if not recursive:
        # => do not recurse into the callees
        return callables_called_by_kernel

    callables_called_by_called_callables = frozenset().union(*(
        callables[clbl_id].get_called_callables(callables)
        for clbl_id in callables_called_by_kernel))
    return callables_called_by_kernel | callables_called_by_called_callables

# }}}


# {{{ get_call_graph

def get_call_graph(t_unit, only_kernel_callables=False):
    """
    Returns a mapping from a callable name to the calls seen in it.

    :arg t_unit: An instance of :class:`TranslationUnit`.
    """
    from constantdict import constantdict

    from loopy.kernel import KernelState

    if t_unit.state < KernelState.CALLS_RESOLVED:
        raise LoopyError("TranslationUnit must have calls resolved in order to"
                         " compute its call graph.")

    knl_callables = frozenset(name for name, clbl in t_unit.callables_table.items()
                              if isinstance(clbl, CallableKernel))

    # stores a mapping from caller -> "direct"" callees
    call_graph = {}

    for name, clbl in t_unit.callables_table.items():
        if (not isinstance(clbl, CallableKernel)
                and only_kernel_callables):
            pass
        else:
            if only_kernel_callables:
                call_graph[name] = (clbl.get_called_callables(t_unit.callables_table,
                                                              recursive=False)
                                    & knl_callables)
            else:
                call_graph[name] = clbl.get_called_callables(t_unit.callables_table,
                                                             recursive=False)

    return constantdict(call_graph)

# }}}


# {{{ get_outer_params

def get_outer_params(domains):
    """
    Returns names of dims that appear only as params in *domains*.

    :arg domains: An instance of :class:`list` of :class:`isl.BasicSet`.
    """
    all_inames = set()
    all_params = set()
    for dom in domains:
        all_inames.update(dom.get_var_names(dim_type.set))
        all_params.update(dom.get_var_names(dim_type.param))

    from loopy.tools import intern_frozenset_of_ids
    return intern_frozenset_of_ids(all_params-all_inames)

# }}}


def get_hw_axis_base_for_codegen(kernel: LoopKernel, iname: str) -> isl.Aff:
    """
    Returns a :class:`isl.PwAff` hardware axes lower bound to serve as an
    offsetting expression
    during the hardware ina
    """
    from loopy.isl_helpers import static_min_of_pw_aff
    from loopy.kernel.data import HardwareConcurrentTag

    assert kernel.iname_tags_of_type(iname, HardwareConcurrentTag)
    bounds = kernel.get_iname_bounds(iname)
    lower_bound = static_min_of_pw_aff(bounds.lower_bound_pw_aff,
                                       constants_only=False)
    return lower_bound

# vim: foldmethod=marker



"""isl helpers"""
from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from warnings import warn

import islpy as isl
from islpy import dim_type

from loopy.diagnostic import LoopyError, StaticValueFindingError


def pw_aff_to_aff(pw_aff):
    if isinstance(pw_aff, isl.Aff):
        return pw_aff

    assert isinstance(pw_aff, isl.PwAff)
    pieces = pw_aff.get_pieces()

    if len(pieces) == 0:
        raise RuntimeError("PwAff does not have any pieces")
    if len(pieces) > 1:
        _, first_aff = pieces[0]
        for _, other_aff in pieces[1:]:
            if not first_aff.plain_is_equal(other_aff):
                raise NotImplementedError("only single-valued piecewise affine "
                        "expressions are supported here--encountered "
                        "multi-valued expression '%s'" % pw_aff)

        return first_aff

    return pieces[0][1]


# {{{ make_slab

def make_slab(space, iname, start, stop, iname_multiplier=1):
    """
    Returns an instance of :class:`islpy._isl.BasicSet`, which satisfies the
    constraint ``start <= iname_multiplier*iname < stop``.

    :arg space: An instance of :class:`islpy._isl.Space`.

    :arg iname:

        Either an instance of :class:`str` as a name of the ``iname`` or a
        tuple of ``(iname_dt, iname_dx)`` indicating the *iname* in the space.

    :arg start:

        An instance of :class:`int`  or an instance of
        :class:`islpy._isl.Aff` indicating the lower bound of
        ``iname_multiplier*iname``(inclusive).

    :arg stop:

        An instance of :class:`int`  or an instance of
        :class:`islpy._isl.Aff` indicating the upper bound of
        ``iname_multiplier*iname``.

    :arg iname_multiplier:

        A strictly positive :class:`int` denoting *iname*'s coefficient in the
        above inequality expression.
    """
    zero = isl.Aff.zero_on_domain(space)

    if isinstance(start, (isl.Aff, isl.PwAff)):
        start, zero = isl.align_two(pw_aff_to_aff(start), zero)
    if isinstance(stop, (isl.Aff, isl.PwAff)):
        stop, zero = isl.align_two(pw_aff_to_aff(stop), zero)

    space = zero.get_domain_space()

    from pymbolic.primitives import ExpressionNode

    from loopy.symbolic import aff_from_expr
    if isinstance(start, ExpressionNode):
        start = aff_from_expr(space, start)
    if isinstance(stop, ExpressionNode):
        stop = aff_from_expr(space, stop)

    if isinstance(start, int):
        start = zero + start
    if isinstance(stop, int):
        stop = zero + stop

    if isinstance(iname, str):
        iname_dt, iname_idx = zero.get_space().get_var_dict()[iname]
    else:
        iname_dt, iname_idx = iname

    iname_aff = zero.add_coefficient_val(iname_dt, iname_idx, 1)

    if iname_multiplier > 0:
        result = (isl.BasicSet.universe(space)
                # start <= iname_multiplier*iname
                .add_constraint(isl.Constraint.inequality_from_aff(
                    iname_multiplier*iname_aff - start))
                # iname_multiplier*iname < stop
                .add_constraint(isl.Constraint.inequality_from_aff(
                    stop-1 - iname_multiplier*iname_aff)))
    else:
        raise LoopyError("iname_multiplier must be strictly positive")

    return result


def make_loop_bounds_from_pwaffs(space, iname, lbound, ubound):
    dt, pos = space.get_var_dict()[iname]
    iname_pwaff = isl.PwAff.var_on_domain(space, dt, pos)

    iname_pwaff, lbound = isl.align_two(iname_pwaff, lbound)
    iname_pwaff, ubound = isl.align_two(iname_pwaff, ubound)
    assert iname_pwaff.space == lbound.space
    assert iname_pwaff.space == ubound.space

    return (
        iname_pwaff.ge_set(lbound)
        &
        iname_pwaff.le_set(ubound))

# }}}


def iname_rel_aff(space, iname, rel, aff):
    """*aff*'s domain space is allowed to not match *space*."""

    dt, pos = space.get_var_dict()[iname]
    assert dt in [isl.dim_type.set, isl.dim_type.param]
    if dt == isl.dim_type.set:
        dt = isl.dim_type.in_

    from islpy import align_spaces
    aff = align_spaces(aff, isl.Aff.zero_on_domain(space))

    if rel in ["==", "<="]:
        return aff.add_coefficient_val(dt, pos, -1)
    elif rel == ">=":
        return aff.neg().add_coefficient_val(dt, pos, 1)
    elif rel == "<":
        return (aff-1).add_coefficient_val(dt, pos, -1)
    elif rel == ">":
        return (aff+1).neg().add_coefficient_val(dt, pos, 1)
    else:
        raise ValueError("unknown value of 'rel': %s" % rel)


# {{{ simplify_pw_aff

def simplify_pw_aff(pw_aff, context=None):
    if context is not None:
        pw_aff = pw_aff.gist_params(context)

    old_pw_aff = pw_aff

    while True:
        restart = False
        did_something = False

        pieces = pw_aff.get_pieces()
        for i, (dom_i, aff_i) in enumerate(pieces):
            for j, (dom_j, aff_j) in enumerate(pieces):
                if i == j:
                    continue

                if aff_i.gist(dom_j).is_equal(aff_j):
                    # aff_i is sufficient to cover aff_j, eliminate aff_j
                    new_pieces = pieces[:]
                    if i < j:
                        new_pieces.pop(j)
                        new_pieces.pop(i)
                    else:
                        new_pieces.pop(i)
                        new_pieces.pop(j)

                    pw_aff = isl.PwAff.alloc(dom_i | dom_j, aff_i)
                    for dom, aff in new_pieces:
                        pw_aff = pw_aff.union_max(isl.PwAff.alloc(dom, aff))

                    restart = True
                    did_something = True
                    break

            if restart:
                break

        if not did_something:
            break

    assert pw_aff.get_aggregate_domain() <= pw_aff.eq_set(old_pw_aff)

    return pw_aff

# }}}


# {{{ static_*_of_pw_aff

def static_extremum_of_pw_aff(pw_aff, constants_only, set_method, what, context):
    if context is not None:
        context = isl.align_spaces(context, pw_aff.get_domain_space(),
                obj_bigger_ok=True).params()
        pw_aff = pw_aff.gist(context)

    pieces = pw_aff.get_pieces()
    if len(pieces) == 1:
        (_, result), = pieces
        if constants_only and not result.is_cst():
            raise StaticValueFindingError("a numeric %s was not found for PwAff '%s'"
                    % (what, pw_aff))
        return result

    from pytools import flatten, memoize

    @memoize
    def is_bounded(set):
        assert set.dim(dim_type.set) == 0
        return (set
                .move_dims(dim_type.set, 0,
                    dim_type.param, 0, set.dim(dim_type.param))
                .is_bounded())

    # put constant bounds with unbounded validity first
    order = [
            (True, False),  # constant, unbounded validity
            (False, False),  # nonconstant, unbounded validity
            (True, True),  # constant, bounded validity
            (False, True),  # nonconstant, bounded validity
            ]

    pieces = flatten([
            [(set, aff) for set, aff in pieces
                if aff.is_cst() == want_is_constant
                and is_bounded(set) == want_is_bounded]
            for want_is_constant, want_is_bounded in order])

    reference = pw_aff.get_aggregate_domain()
    if context is not None:
        reference = reference.intersect(context)

    # {{{ find bounds that are also global bounds

    for set, candidate_aff in pieces:
        # gist can be time-consuming, try without first
        for use_gist in [False, True]:
            if use_gist:
                candidate_aff = candidate_aff.gist(set)

            if constants_only and not candidate_aff.is_cst():
                continue

            if reference <= set_method(pw_aff, candidate_aff):
                return candidate_aff

    # }}}

    raise StaticValueFindingError("a static %s was not found for PwAff '%s'"
            % (what, pw_aff))


def static_min_of_pw_aff(pw_aff, constants_only, context=None):
    return static_extremum_of_pw_aff(pw_aff, constants_only, isl.PwAff.ge_set,
            "minimum", context)


def static_max_of_pw_aff(pw_aff, constants_only, context=None):
    return static_extremum_of_pw_aff(pw_aff, constants_only, isl.PwAff.le_set,
            "maximum", context)


def static_value_of_pw_aff(pw_aff, constants_only, context=None):
    return static_extremum_of_pw_aff(pw_aff, constants_only, isl.PwAff.eq_set,
            "value", context)

# }}}


# {{{ duplicate_axes

def duplicate_axes(isl_obj, duplicate_inames, new_inames):
    """
    Duplicates dim names in *duplicate_inames* with corresponding names in
    *new_inames*.

    .. testsetup::

        >>> import islpy as isl
        >>> from loopy.isl_helpers import duplicate_axes

    .. doctest::

        >>> bset = isl.BasicSet("{[i, j]: 0<=i<10 and 0<=j<30}")
        >>> duplicate_axes(bset, ("i",), ("i'",))
        BasicSet("{ [i, j, i'] : 0 <= i <= 9 and 0 <= j <= 29 and 0 <= i' <= 9 }")
    """
    if isinstance(isl_obj, list):
        return [
                duplicate_axes(i, duplicate_inames, new_inames)
                for i in isl_obj]

    if not duplicate_inames:
        return isl_obj

    def _align_and_intersect(d1, d2):
        d1, d2 = isl.align_two(d1, d2)
        return d1 & d2

    old_name_to_new_name = dict(zip(duplicate_inames, new_inames))

    dup_isl_obj = isl_obj

    for old_name, (dt, pos) in isl_obj.get_var_dict().items():
        dup_isl_obj = dup_isl_obj.set_dim_name(dt, pos,
                                               old_name_to_new_name.get(old_name,
                                                                        old_name))

    return _align_and_intersect(dup_isl_obj, isl_obj)

# }}}


def is_nonnegative(expr, over_set):
    from pymbolic.primitives import Product

    from loopy.symbolic import aff_from_expr

    if isinstance(expr, Product) and all(
            is_nonnegative(child, over_set) for child in expr.children):
        return True

    space = over_set.get_space()
    try:
        aff = aff_from_expr(space, -expr-1)
    except Exception:
        return None
    expr_neg_set = isl.BasicSet.universe(space).add_constraint(
            isl.Constraint.inequality_from_aff(aff))

    return over_set.intersect(expr_neg_set).is_empty()


# {{{ convexify

def convexify(domain):
    """Try a few ways to get *domain* to be a BasicSet, i.e.
    explicitly convex.
    """

    if isinstance(domain, isl.BasicSet):
        return domain

    dom_bsets = domain.get_basic_sets()
    if len(dom_bsets) == 1:
        domain, = dom_bsets
        return domain

    hull_domain = domain.simple_hull()
    if isl.Set.from_basic_set(hull_domain) <= domain:
        return hull_domain

    domain = domain.coalesce()

    dom_bsets = domain.get_basic_sets()
    if len(domain.get_basic_sets()) == 1:
        domain, = dom_bsets
        return domain

    hull_domain = domain.simple_hull()
    if isl.Set.from_basic_set(hull_domain) <= domain:
        return hull_domain

    dom_bsets = domain.get_basic_sets()
    assert len(dom_bsets) > 1

    print("PIECES:")
    for dbs in dom_bsets:
        print("  %s" % (isl.Set.from_basic_set(dbs).gist(domain)))
    raise NotImplementedError("Could not find convex representation of set")

# }}}


# {{{ boxify

def boxify(cache_manager, domain, box_inames, context):
    var_dict = domain.get_var_dict(dim_type.set)
    box_iname_indices = [var_dict[iname][1] for iname in box_inames]
    n_nonbox_inames = min(box_iname_indices)

    assert box_iname_indices == list(range(
            n_nonbox_inames, domain.dim(dim_type.set)))

    n_old_parameters = domain.dim(dim_type.param)
    domain = domain.move_dims(
            dim_type.param, n_old_parameters, dim_type.set, 0, n_nonbox_inames)

    result = domain
    zero = isl.Aff.zero_on_domain(result.space)

    for i in range(len(box_iname_indices)):
        result = result.eliminate(dim_type.set, i, 1)

        iname_aff = zero.add_coefficient_val(dim_type.in_, i, 1)

        def add_in_dims(aff):
            return aff.add_dims(dim_type.in_, len(box_inames))

        iname_min = add_in_dims(cache_manager.dim_min(domain, i)).coalesce()
        iname_max = add_in_dims(cache_manager.dim_max(domain, i)).coalesce()

        iname_slab = (iname_min.le_set(iname_aff)
                .intersect(iname_max.ge_set(iname_aff)))

        for i, iname in enumerate(box_inames):
            iname_slab = iname_slab.set_dim_name(dim_type.set, i, iname)

        if context is not None:
            iname_slab, context = isl.align_two(iname_slab, context)
            iname_slab = iname_slab.gist(context)
        iname_slab = iname_slab.coalesce()

        result = result & iname_slab

    result = result.move_dims(
            dim_type.set, 0, dim_type.param, n_old_parameters, n_nonbox_inames)

    return convexify(result)

# }}}


def project_out(set, inames):
    for iname in inames:
        var_dict = set.get_var_dict()
        dt, dim_idx = var_dict[iname]
        set = set.project_out(dt, dim_idx, 1)

    return set


def obj_involves_variable(obj, var_name):
    loc = obj.get_var_dict().get(var_name)
    if loc is not None:
        if not obj.get_coefficient_val(*loc).is_zero():
            return True

    for idiv in obj.dim(dim_type.div):
        if obj_involves_variable(obj.get_div(idiv), var_name):
            return True

    return False


# {{{ get_simple_strides

def get_simple_strides(bset, key_by="name"):
    """Return a dictionary from inames to strides in bset. Each stride is
    returned as a :class:`islpy.Val`. If no stride can be determined, the
    corresponding key will not be present in the returned dictionary.

    This only recognizes simple strides involving single variables.

    :arg key_by: "index" or "name"
    """
    result = {}

    comp_div_set_pieces = convexify(bset.compute_divs()).get_basic_sets()
    assert len(comp_div_set_pieces) == 1
    bset, = comp_div_set_pieces

    def _get_indices_and_coeffs(obj, dts):
        result = []
        for dt in dts:
            for dim_idx in range(obj.dim(dt)):
                coeff_val = obj.get_coefficient_val(dt, dim_idx)
                if not coeff_val.is_zero():
                    result.append((dt, dim_idx, coeff_val))

        return result

    for cns in bset.get_constraints():
        if not cns.is_equality():
            continue
        aff = cns.get_aff()

        # recognizes constraints of the form
        #  -i0 + 2*floor((i0)/2) == 0

        divs_with_coeffs = _get_indices_and_coeffs(aff, [dim_type.div])
        if len(divs_with_coeffs) != 1:
            continue

        (_, idiv, div_coeff), = divs_with_coeffs

        div = aff.get_div(idiv)

        # check for sub-divs
        if _get_indices_and_coeffs(div, [dim_type.div]):
            # found one -> not supported
            continue

        denom = div.get_denominator_val().to_python()

        # if the coefficient in front of the div is not the same as the denominator
        if not div_coeff.div(denom).is_one():
            # not supported
            continue

        inames_and_coeffs = _get_indices_and_coeffs(
                div, [dim_type.param, dim_type.in_])

        if len(inames_and_coeffs) != 1:
            continue

        (dt, dim_idx, coeff), = inames_and_coeffs

        if not (coeff * denom).is_one():
            # not supported
            continue

        inames_and_coeffs = _get_indices_and_coeffs(
                aff, [dim_type.param, dim_type.in_])

        if len(inames_and_coeffs) != 1:
            continue

        (outer_dt, outer_dim_idx, outer_coeff), = inames_and_coeffs
        if (not outer_coeff.neg().is_one()
                or (outer_dt, outer_dim_idx) != (dt, dim_idx)):
            # not supported
            continue

        if key_by == "name":
            key = bset.get_dim_name(dt, dim_idx)
        elif key_by == "index":
            key_dt = dt if dt != dim_type.in_ else dim_type.set

            key = (key_dt, dim_idx)
        else:
            raise ValueError("invalid value of 'key_by")

        result[key] = denom

    return result

# }}}


# {{{ find_max_of_pwaff_with_params

def find_max_of_pwaff_with_params(pw_aff, n_allowed_params):
    if n_allowed_params is None:
        return pw_aff

    extra_dim_idx = pw_aff.dim(dim_type.param,)
    pw_aff = pw_aff.add_dims(dim_type.param, 1)

    zero = isl.Aff.zero_on_domain(pw_aff.domain().space)
    extra_dim = zero.set_coefficient_val(dim_type.param, extra_dim_idx, 1)

    pw_aff_set = pw_aff.eq_set(extra_dim)

    pw_aff_set = pw_aff_set.move_dims(
            dim_type.set, 0, dim_type.param, n_allowed_params,
            pw_aff_set.dim(dim_type.param) - n_allowed_params)

    return pw_aff_set.dim_max(pw_aff_set.dim(dim_type.set)-1)

# }}}


# {{{ subst_into_pw(qpolynomial|aff)

def set_dim_name(obj, dt, pos, name):
    assert isinstance(name, str)
    if isinstance(obj, (isl.PwQPolynomial, isl.BasicSet)):
        return obj.set_dim_name(dt, pos, name)
    elif isinstance(obj, isl.PwAff):
        # work around missing isl_pw_aff_set_dim_name for now.
        # https://github.com/inducer/loopy/pull/233/files#r580594032
        return obj.set_dim_id(dt, pos, isl.Id.read_from_str(obj.get_ctx(), name))
    else:
        raise NotImplementedError(f"not implemented for {type(obj)}.")


def get_param_subst_domain(new_space, base_obj, subst_dict):
    """Modify the :mod:`islpy` object *base_obj* to incorporate parameters for
    the keys of *subst_dict*, and rename existing parameters to include a
    trailing prime.

    :arg new_space: A :class:`islpy.Space` for that contains the keys of
        *subst_dict*
    :arg subst_dict: A dictionary mapping parameters occurring in *base_obj*
        to their values in terms of variables in *new_space*
    :returns: a tuple ``(base_obj, subst_domain, subst_dict)``, where
        *base_obj* is the passed *base_obj* with the space extended to cover
        the new parameters in *new_space*, *subst_domain* is an
        :class:`islpy.BasicSet` incorporating the constraints from *subst_dict*
        and existing in the same space as *base_obj*, and *subst_dict*
        is a copy of the passed *subst_dict* modified to incorporate primed
        variable names in the keys.
    """

    # {{{ rename subst_dict keys and base_obj parameters to include trailing prime

    i_begin_subst_space = base_obj.dim(dim_type.param)

    new_subst_dict = {}
    for i in range(i_begin_subst_space):
        old_name = base_obj.space.get_dim_name(dim_type.param, i)
        new_name = old_name + "'"
        new_subst_dict[new_name] = subst_dict[old_name]
        base_obj = set_dim_name(base_obj, dim_type.param, i, new_name)

    subst_dict = new_subst_dict
    del new_subst_dict

    # }}}

    # {{{ add dimensions to base_obj

    base_obj = base_obj.add_dims(dim_type.param, new_space.dim(dim_type.param))
    for i in range(new_space.dim(dim_type.param)):
        base_obj = set_dim_name(base_obj, dim_type.param, i+i_begin_subst_space,
                new_space.get_dim_name(dim_type.param, i))

    # }}}

    # {{{ build subst_domain

    subst_domain = isl.BasicSet.universe(base_obj.space).params()

    from loopy.symbolic import guarded_aff_from_expr
    for i in range(i_begin_subst_space):
        name = base_obj.space.get_dim_name(dim_type.param, i)
        aff = guarded_aff_from_expr(subst_domain.space, subst_dict[name])
        aff = aff.set_coefficient_val(dim_type.param, i, -1)
        subst_domain = subst_domain.add_constraint(
                isl.Constraint.equality_from_aff(aff))

    # }}}

    return base_obj, subst_domain, subst_dict


def subst_into_pwqpolynomial(new_space, poly, subst_dict):
    """
    Returns an instance of :class:`islpy.PwQPolynomial` with substitutions from
    *subst_dict* substituted into *poly*.

    :arg poly: an instance of :class:`islpy.PwQPolynomial`
    :arg subst_dict: a mapping from parameters of *poly* to
        :class:`pymbolic.primitives.Expression` made up of terms comprising the
        parameters of *new_space*. The expression must be affine in the param
        dims of *new_space*.
    """
    if not poly.get_pieces():
        # pw poly is univserally zero
        result = isl.PwQPolynomial.zero(new_space.insert_dims(dim_type.out, 0, 1))
        assert result.dim(dim_type.out) == 1
        return result

    i_begin_subst_space = poly.dim(dim_type.param)

    poly, subst_domain, subst_dict = get_param_subst_domain(
            new_space, poly, subst_dict)

    from loopy.symbolic import qpolynomial_from_expr, qpolynomial_to_expr
    new_pieces = []
    for valid_set, qpoly in poly.get_pieces():
        valid_set = valid_set & subst_domain
        if valid_set.plain_is_empty():
            continue

        valid_set = valid_set.project_out(dim_type.param, 0, i_begin_subst_space)
        from pymbolic.mapper.substitutor import SubstitutionMapper, make_subst_func
        sub_mapper = SubstitutionMapper(make_subst_func(subst_dict))
        expr = sub_mapper(qpolynomial_to_expr(qpoly))
        qpoly = qpolynomial_from_expr(valid_set.space, expr)

        new_pieces.append((valid_set, qpoly))

    if not new_pieces:
        raise ValueError("no pieces of PwQPolynomial survived the substitution")

    valid_set, qpoly = new_pieces[0]
    result = isl.PwQPolynomial.alloc(valid_set, qpoly)
    for valid_set, qpoly in new_pieces[1:]:
        result = result.add_disjoint(
                isl.PwQPolynomial.alloc(valid_set, qpoly))

    assert result.dim(dim_type.out)
    return result


def subst_into_pwaff(new_space, pwaff, subst_dict):
    """
    Returns an instance of :class:`islpy.PwAff` with substitutions from
    *subst_dict* substituted into *pwaff*.

    :arg pwaff: an instance of :class:`islpy.PwAff`
    :arg subst_dict: a mapping from parameters of *pwaff* to
        :class:`pymbolic.primitives.Expression` made up of terms comprising the
        parameters of *new_space*. The expression must be affine in the param
        dims of *new_space*.
    """
    from functools import reduce

    from pymbolic.mapper.substitutor import SubstitutionMapper, make_subst_func

    from loopy.symbolic import aff_from_expr, aff_to_expr

    i_begin_subst_space = pwaff.dim(dim_type.param)
    pwaff, subst_domain, subst_dict = get_param_subst_domain(
            new_space, pwaff, subst_dict)
    subst_mapper = SubstitutionMapper(make_subst_func(subst_dict))
    pwaffs = []

    for valid_set, qpoly in pwaff.get_pieces():
        valid_set = valid_set & subst_domain
        if valid_set.plain_is_empty():
            continue

        valid_set = valid_set.project_out(dim_type.param, 0, i_begin_subst_space)
        aff = aff_from_expr(valid_set.space, subst_mapper(aff_to_expr(qpoly)))

        pwaffs.append(isl.PwAff.alloc(valid_set, aff))

    if not pwaffs:
        raise ValueError("no pieces of PwAff survived the substitution")

    return reduce(lambda pwaff1, pwaff2: pwaff1.union_add(pwaff2),
                  pwaffs).coalesce()

# }}}


# {{{ add_and_name_dims

def add_and_name_dims(isl_obj, dt, names):
    """Append dimensions of the specified dimension type to the provided ISL
    object, and set their names.

    :arg isl_obj: An :class:`islpy.Set` or  :class:`islpy.Map` to which
        new dimensions will be added.

    :arg dt: An :class:`islpy.dim_type`, i.e., an :class:`int`, specifying the
        dimension type for the new dimensions.

    :arg names: An iterable of :class:`str` values specifying the names of the
        new dimensions to be added.

    :returns: An object of the same type as *isl_obj* with the new dimensions
        added and named.

    """

    new_idx_start = isl_obj.dim(dt)
    isl_obj = isl_obj.add_dims(dt, len(names))
    for i, name in enumerate(names):
        isl_obj = isl_obj.set_dim_name(dt, new_idx_start+i, name)
    return isl_obj

# }}}


# {{{ add_eq_constraint_from_names

def add_eq_constraint_from_names(isl_obj, var1, var2):
    """Add constraint *var1* = *var2* to an ISL object.

    :arg isl_obj: An :class:`islpy.Set` or  :class:`islpy.Map` to which
        a new constraint will be added.

    :arg var1: A :class:`str` specifying the name of the first variable
        involved in constraint *var1* = *var2*.

    :arg var2: A :class:`str` specifying the name of the second variable
        involved in constraint *var1* = *var2*.

    :returns: An object of the same type as *isl_obj* with the constraint
        *var1* = *var2*.

    """
    return isl_obj.add_constraint(
               isl.Constraint.eq_from_names(
                   isl_obj.space,
                   {1: 0, var1: 1, var2: -1}))

# }}}


# {{{ find_and_rename_dim

def find_and_rename_dim(isl_obj, dt, old_name, new_name):
    """Rename a dimension in an ISL object.

    :arg isl_obj: An :class:`islpy.Set` or  :class:`islpy.Map` containing the
        dimension to be renamed.

    :arg dt: An :class:`islpy.dim_type` (i.e., :class:`int`) specifying the
        dimension type containing the dimension to be renamed.

    :arg old_name: A :class:`str` specifying the name of the dimension to be
        renamed.

    :arg new_name: A :class:`str` specifying the new name of the dimension to
        be renamed.

    :returns: An object of the same type as *isl_obj* with the dimension
        *old_name* renamed to *new_name*.

    """
    return isl_obj.set_dim_name(
            dt, isl_obj.find_dim_by_name(dt, old_name), new_name)

# }}}


def simplify_via_aff(expr):
    warn("simplify_via_aff has moved to loopy.symbolic. "
            "Importing it from loopy.isl_helpers will stop working in July 2022.",
            DeprecationWarning, stacklevel=2)
    from loopy.symbolic import simplify_via_aff
    return simplify_via_aff(expr)

# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from typing import TYPE_CHECKING

import numpy as np
from constantdict import constantdict

from loopy.diagnostic import LoopyError
from loopy.kernel.function_interface import ScalarCallable
from loopy.types import NumpyType


if TYPE_CHECKING:
    from loopy.translation_unit import CallablesTable


class MakeTupleCallable(ScalarCallable):
    def with_types(self, arg_id_to_dtype, callables_table):
        new_arg_id_to_dtype = constantdict(arg_id_to_dtype).mutate()
        for i in range(len(arg_id_to_dtype)):
            if i in arg_id_to_dtype and arg_id_to_dtype[i] is not None:
                new_arg_id_to_dtype[-i-1] = new_arg_id_to_dtype[i]

        return (self.copy(arg_id_to_dtype=new_arg_id_to_dtype.finish(),
                          name_in_target="loopy_make_tuple"),
                callables_table)

    def with_descrs(self, arg_id_to_descr, callables_table):
        from loopy.kernel.function_interface import ValueArgDescriptor
        new_arg_id_to_descr = {(id, ValueArgDescriptor()):
            (-id-1, ValueArgDescriptor()) for id in arg_id_to_descr}

        return (
                self.copy(arg_id_to_descr=constantdict(new_arg_id_to_descr)),
                callables_table)


class IndexOfCallable(ScalarCallable):
    def with_types(self, arg_id_to_dtype, callables_table):
        new_arg_id_to_dtype = {i: dtype
                               for i, dtype in arg_id_to_dtype.items()
                               if dtype is not None}
        new_arg_id_to_dtype[-1] = NumpyType(np.int32)

        return (self.copy(arg_id_to_dtype=constantdict(new_arg_id_to_dtype)),
                callables_table)

    def emit_call(self, expression_to_code_mapper, expression, target):
        from pymbolic.primitives import Subscript

        if len(expression.parameters) != 1:
            raise LoopyError("%s takes exactly one argument" % self.name)
        arg, = expression.parameters
        if not isinstance(arg, Subscript):
            raise LoopyError(
                    "argument to %s must be a subscript" % self.name)

        ary = expression_to_code_mapper.find_array(arg)

        from pymbolic import evaluate

        from loopy.kernel.array import get_access_info
        access_info = get_access_info(expression_to_code_mapper.kernel,
                ary, arg.index, lambda expr: evaluate(expr,
                    expression_to_code_mapper.codegen_state.var_subst_map),
                expression_to_code_mapper.codegen_state.vectorization_info)

        from loopy.kernel.data import ImageArg
        if isinstance(ary, ImageArg):
            raise LoopyError("%s does not support images" % self.name)

        if self.name == "indexof":
            return access_info.subscripts[0]
        elif self.name == "indexof_vec":
            from loopy.kernel.array import VectorArrayDimTag
            ivec = None
            for iaxis, dim_tag in enumerate(ary.dim_tags):
                if isinstance(dim_tag, VectorArrayDimTag):
                    ivec = iaxis

            if ivec is None:
                return access_info.subscripts[0]
            else:
                return (
                    access_info.subscripts[0]*ary.shape[ivec]
                    + access_info.vector_index)

        else:
            raise RuntimeError("should not get here")

    def emit_call_insn(self, insn, target, expression_to_code_mapper):
        return self.emit_call(
                expression_to_code_mapper,
                insn.expression,
                target), True


def get_loopy_callables() -> CallablesTable:
    """
    Returns a mapping from function ids to corresponding
    :class:`loopy.kernel.function_interface.InKernelCallable` for functions
    whose interface is provided by :mod:`loopy`. Callables that fall in this
    category are --

    - reductions leading to function calls like ``argmin``, ``argmax``.
    - callables that have a predefined meaning in :mod:`loo.py` like
      ``make_tuple``, ``index_of``, ``indexof_vec``.
    """
    return {
            "make_tuple": MakeTupleCallable(name="make_tuple"),
            "indexof": IndexOfCallable(name="indexof"),
            "indexof_vec": IndexOfCallable(name="indexof_vec"),
            }


# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import islpy as isl

from loopy.translation_unit import for_each_kernel


def potential_loop_nest_map(kernel):
    """Returns a dictionary mapping inames to other inames that *could*
    be nested around them.

    * :seealso: :func:`loopy.schedule.loop_nest_map`
    * :seealso: :func:`loopy.schedule.find_loop_nest_around_map`
    """

    result = {}

    all_inames = kernel.all_inames()
    iname_to_insns = kernel.iname_to_insns()

    # examine pairs of all inames--O(n**2), I know.
    for inner_iname in all_inames:
        inner_result = set()
        for outer_iname in all_inames:
            if inner_iname == outer_iname:
                continue

            if iname_to_insns[inner_iname] <= iname_to_insns[outer_iname]:
                inner_result.add(outer_iname)

        if inner_result:
            result[inner_iname] = inner_result

    return result


@for_each_kernel
def merge_loop_domains(kernel):
    # FIXME: This should be moved to loopy.transforms.iname
    from loopy.kernel.tools import is_domain_dependent_on_inames

    while True:
        lnm = potential_loop_nest_map(kernel)
        parents_per_domain = kernel.parents_per_domain()
        all_parents_per_domain = kernel.all_parents_per_domain()

        iname_to_insns = kernel.iname_to_insns()

        new_domains = None

        for inner_iname, outer_inames in lnm.items():
            for outer_iname in outer_inames:
                # {{{ check if it's safe to merge

                inner_domain_idx = kernel.get_home_domain_index(inner_iname)
                outer_domain_idx = kernel.get_home_domain_index(outer_iname)

                if inner_domain_idx == outer_domain_idx:
                    break

                if (not iname_to_insns[inner_iname]
                        or not iname_to_insns[outer_iname]):
                    # Inames without instructions occur when used in
                    # a SubArrayRef. We don't want monster SubArrayRef domains,
                    # so refuse to merge those.
                    continue

                if iname_to_insns[inner_iname] != iname_to_insns[outer_iname]:
                    # The two inames are imperfectly nested. Domain fusion
                    # might be invalid when the inner loop is empty, leading to
                    # the outer loop also being empty.

                    # FIXME: Not fully correct, does not consider reductions
                    # https://gitlab.tiker.net/inducer/loopy/issues/172
                    continue

                if (
                        outer_domain_idx in all_parents_per_domain[inner_domain_idx]
                        and not
                        outer_domain_idx == parents_per_domain[inner_domain_idx]):
                    # Outer domain is not a direct parent of the inner
                    # domain. Unable to merge.
                    continue

                outer_dom = kernel.domains[outer_domain_idx]
                inner_dom = kernel.domains[inner_domain_idx]

                outer_inames = set(outer_dom.get_var_names(isl.dim_type.set))
                if is_domain_dependent_on_inames(kernel, inner_domain_idx,
                        outer_inames):
                    # Bounds of inner domain depend on outer domain.
                    # Unable to merge.
                    continue

                # }}}

                new_domains = kernel.domains[:]
                min_idx = min(inner_domain_idx, outer_domain_idx)
                max_idx = max(inner_domain_idx, outer_domain_idx)

                del new_domains[max_idx]
                del new_domains[min_idx]

                outer_dom, inner_dom = isl.align_two(outer_dom, inner_dom)

                new_domains.insert(min_idx, inner_dom & outer_dom)
                break

            if new_domains:
                break

        if not new_domains:
            # Nothing was accomplished in the last loop trip, time to quit.
            break

        kernel = kernel.copy(domains=new_domains)

    return kernel


# vim: fdm=marker



from __future__ import annotations


__copyright__ = """
Copyright (C) 2016 Matt Wala
Copyright (C) 2020 University of Illinois Board of Trustees
Copyright (C) 2022 Kaushik Kulkarni
"""

__doc__ = """
.. autofunction:: get_block_boundaries
.. autofunction:: temporaries_read_in_subkernel
.. autofunction:: args_read_in_subkernel
.. autofunction:: args_written_in_subkernel
.. autofunction:: supporting_temporary_names

.. autoclass:: KernelArgInfo
.. autoclass:: SubKernelArgInfo

.. autofunction:: get_kernel_arg_info
.. autofunction:: get_subkernel_arg_info

.. autofunction:: get_return_from_kernel_mapping

.. autoclass:: AccessMapDescriptor
.. autoclass:: WriteRaceChecker

.. autoclass:: LoopNestTree
.. autoclass:: LoopTree

.. autofunction:: separate_loop_nest
.. autofunction:: get_partial_loop_nest_tree
.. autofunction:: get_loop_tree

References
^^^^^^^^^^

.. class:: InameStrSet

    See :class:`loopy.typing.InameStrSet`
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import enum
import itertools
from dataclasses import dataclass
from functools import cached_property, reduce
from typing import TYPE_CHECKING, AbstractSet, Sequence

from constantdict import constantdict
from typing_extensions import TypeAlias

import islpy as isl
from pytools import memoize_method, memoize_on_first_arg

from loopy.diagnostic import LoopyError
from loopy.kernel.data import AddressSpace, ArrayArg, TemporaryVariable
from loopy.schedule.tree import Tree
from loopy.typing import InameStr, InameStrSet, not_none


if TYPE_CHECKING:
    from collections.abc import Callable, Collection, Mapping

    from loopy.kernel import LoopKernel
    from loopy.schedule import ScheduleItem


# {{{ block boundary finder

def get_block_boundaries(schedule: Sequence[ScheduleItem]) -> Mapping[int, int]:
    r"""
    Return a dictionary mapping indices of
    :class:`loopy.schedule.BeginBlockItem`\ s to
    :class:`loopy.schedule.EndBlockItem`\ s and vice versa.
    """
    from loopy.schedule import BeginBlockItem, EndBlockItem
    block_bounds = {}
    active_blocks = []
    for idx, sched_item in enumerate(schedule):
        if isinstance(sched_item, BeginBlockItem):
            active_blocks.append(idx)
        elif isinstance(sched_item, EndBlockItem):
            start = active_blocks.pop()
            block_bounds[start] = idx
            block_bounds[idx] = start
    return block_bounds

# }}}


# {{{ subkernel tools

def temporaries_read_in_subkernel(
        kernel: LoopKernel, subkernel_name: str) -> frozenset[str]:
    from loopy.kernel.tools import get_subkernel_to_insn_id_map
    insn_ids = get_subkernel_to_insn_id_map(kernel)[subkernel_name]
    inames = frozenset().union(*(kernel.insn_inames(insn_id)
                                 for insn_id in insn_ids))
    domain_idxs = {kernel.get_home_domain_index(iname) for iname in inames}
    params = frozenset().union(*(
        kernel.domains[dom_idx].get_var_names(isl.dim_type.param)
        for dom_idx in domain_idxs))

    return (frozenset(tv
                for insn_id in insn_ids
                for tv in kernel.id_to_insn[insn_id].read_dependency_names()
                if tv in kernel.temporary_variables)
            | (params & frozenset(kernel.temporary_variables)))


def temporaries_written_in_subkernel(
        kernel: LoopKernel, subkernel_name: str) -> frozenset[str]:
    from loopy.kernel.tools import get_subkernel_to_insn_id_map
    insn_ids = get_subkernel_to_insn_id_map(kernel)[subkernel_name]
    return frozenset(tv
                for insn_id in insn_ids
                for tv in kernel.id_to_insn[insn_id].assignee_var_names()
                if tv in kernel.temporary_variables)


def args_read_in_subkernel(
        kernel: LoopKernel, subkernel_name: str) -> frozenset[str]:
    from loopy.kernel.tools import get_subkernel_to_insn_id_map
    insn_ids = get_subkernel_to_insn_id_map(kernel)[subkernel_name]
    inames = frozenset().union(*(kernel.insn_inames(insn_id)
                                 for insn_id in insn_ids))
    domain_idxs = {kernel.get_home_domain_index(iname) for iname in inames}
    params = frozenset().union(*(
        kernel.domains[dom_idx].get_var_names(isl.dim_type.param)
        for dom_idx in domain_idxs))
    return (frozenset(arg
                for insn_id in insn_ids
                for arg in kernel.id_to_insn[insn_id].read_dependency_names()
                if arg in kernel.arg_dict)
            | (params & frozenset(kernel.arg_dict)))


def args_written_in_subkernel(
        kernel: LoopKernel, subkernel_name: str) -> frozenset[str]:
    from loopy.kernel.tools import get_subkernel_to_insn_id_map
    insn_ids = get_subkernel_to_insn_id_map(kernel)[subkernel_name]
    return frozenset(arg
                for insn_id in insn_ids
                for arg in kernel.id_to_insn[insn_id].assignee_var_names()
                if arg in kernel.arg_dict)


def supporting_temporary_names(
        kernel: LoopKernel, tv_names: frozenset[str]) -> frozenset[str]:
    result: set[str] = set()

    for name in tv_names:
        tv = kernel.temporary_variables[name]
        for supp_name in tv.supporting_names():
            if supp_name in kernel.temporary_variables:
                result.add(supp_name)

    return frozenset(result)

# }}}


# {{{ argument lists

@dataclass(frozen=True)
class KernelArgInfo:
    """
    .. autoattribute:: passed_arg_names
    .. autoattribute:: written_names
    """

    passed_arg_names: Sequence[str]
    written_names: frozenset[str]

    @property
    def passed_names(self) -> Sequence[str]:
        return self.passed_arg_names


@dataclass(frozen=True)
class SubKernelArgInfo(KernelArgInfo):
    """Inherits from :class:`KernelArgInfo`.

    .. autoattribute:: passed_inames
    .. autoattribute:: passed_temporaries
    """

    passed_inames: Sequence[str]
    passed_temporaries: Sequence[str]

    @property
    def passed_names(self) -> Sequence[str]:
        return (list(self.passed_arg_names)
                + list(self.passed_inames)
                + list(self.passed_temporaries))


def _should_temp_var_be_passed(tv: TemporaryVariable) -> bool:
    return tv.address_space == AddressSpace.GLOBAL and tv.initializer is None


class _SupportingNameTracker:
    def __init__(self, kernel: LoopKernel):
        self.kernel = kernel
        self.name_to_main_name: dict[str, str] = {}

    def add_supporting_names_for(self, name):
        var_descr = self.kernel.get_var_descriptor(name)
        for supp_name in var_descr.supporting_names():
            self.name_to_main_name[supp_name] = (
                    self.name_to_main_name.get(supp_name, frozenset())
                    | {name})

    def get_additional_args_and_tvs(
            self, already_passed: set[str]
            ) -> tuple[list[str], list[str]]:
        additional_args = []
        additional_temporaries = []

        for supporting_name in sorted(frozenset(self.name_to_main_name)):
            if supporting_name not in already_passed:
                already_passed.add(supporting_name)
                var_descr = self.kernel.get_var_descriptor(supporting_name)
                if isinstance(var_descr, TemporaryVariable):
                    if _should_temp_var_be_passed(var_descr):
                        additional_temporaries.append(supporting_name)
                else:
                    additional_args.append(supporting_name)

        return additional_args, additional_temporaries


def _process_args_for_arg_info(
        kernel: LoopKernel, args_read: set[str], args_written: set[str],
        supp_name_tracker: _SupportingNameTracker, used_only: bool,
        ) -> list[str]:

    args_expected: set[str] = set()

    passed_arg_names = []
    for arg in kernel.args:
        if used_only and not (arg.name in args_read or arg.name in args_written):
            continue

        try:
            args_expected.remove(arg.name)
        except KeyError:
            pass

        # Disregard the original array if it had a sep-tagged axis.
        if isinstance(arg, ArrayArg):
            if not arg._separation_info:
                passed_arg_names.append(arg.name)
                supp_name_tracker.add_supporting_names_for(arg.name)
            else:
                for sep_name in sorted(arg._separation_info.subarray_names.values()):
                    # Separated arrays occur later in the argument list.
                    # Mark them as accessed if the original array was,
                    # we'll stumble on them when it is their turn.
                    # Add them to args_expected to ensure they're not missed.
                    if arg.name in args_read:
                        args_read.add(sep_name)
                        args_expected.add(sep_name)
                    if arg.name in args_written:
                        args_written.add(sep_name)
                        args_expected.add(sep_name)

        else:
            passed_arg_names.append(arg.name)
            supp_name_tracker.add_supporting_names_for(arg.name)

    assert not args_expected

    return passed_arg_names


def get_kernel_arg_info(kernel: LoopKernel) -> KernelArgInfo:
    args_written = set(kernel.arg_dict) & kernel.get_written_variables()

    supp_name_tracker = _SupportingNameTracker(kernel)

    passed_arg_names = _process_args_for_arg_info(kernel,
            args_read=set(), args_written=args_written,
            supp_name_tracker=supp_name_tracker,
            used_only=False)

    additional_args, additional_temporaries = \
            supp_name_tracker.get_additional_args_and_tvs(
                    already_passed=(
                        set(passed_arg_names)))

    assert not additional_temporaries

    return KernelArgInfo(
            passed_arg_names=passed_arg_names + additional_args,
            written_names=frozenset(args_written))


def get_subkernel_arg_info(
        kernel: LoopKernel, subkernel_name: str) -> SubKernelArgInfo:
    assert kernel.linearization is not None

    args_read = set(args_read_in_subkernel(kernel, subkernel_name))
    args_written = set(args_written_in_subkernel(kernel, subkernel_name))

    tvs_read = temporaries_read_in_subkernel(kernel, subkernel_name)
    tvs_written = set(temporaries_written_in_subkernel(kernel, subkernel_name))

    supp_name_tracker = _SupportingNameTracker(kernel)

    passed_arg_names = _process_args_for_arg_info(kernel,
            args_read=args_read, args_written=args_written,
            supp_name_tracker=supp_name_tracker,
            used_only=True)

    passed_temporaries: list[str] = []
    for tv_name in sorted(tvs_read | tvs_written):
        supp_name_tracker.add_supporting_names_for(tv_name)
        tv = kernel.temporary_variables[tv_name]

        if _should_temp_var_be_passed(tv):
            if tv.base_storage:
                if tv_name in tvs_written:
                    if tv_name in tvs_written:
                        tvs_written.add(tv.base_storage)
            else:
                passed_temporaries.append(tv.name)

    additional_args, additional_temporaries = \
            supp_name_tracker.get_additional_args_and_tvs(
                    already_passed=(
                        set(passed_arg_names) | set(passed_temporaries)))

    from loopy.kernel.tools import get_subkernel_extra_inames

    return SubKernelArgInfo(
            passed_arg_names=passed_arg_names + additional_args,
            passed_inames=sorted(get_subkernel_extra_inames(kernel)[subkernel_name]),
            passed_temporaries=passed_temporaries + additional_temporaries,
            written_names=frozenset(args_written | tvs_written))

# }}}


# {{{ get_return_from_kernel_mapping

def get_return_from_kernel_mapping(kernel: LoopKernel) -> Mapping[int, int | None]:
    """
    Returns a mapping from schedule index of every schedule item (S) in
    *kernel* to the schedule index of :class:`loopy.schedule.ReturnFromKernel`
    of the active sub-kernel at 'S'.
    """
    from loopy.kernel import LoopKernel
    from loopy.schedule import (
        Barrier,
        CallKernel,
        EnterLoop,
        LeaveLoop,
        ReturnFromKernel,
        RunInstruction,
    )
    assert isinstance(kernel, LoopKernel)
    assert isinstance(kernel.linearization, list)
    return_from_kernel_idxs: dict[int, int | None] = {}
    current_return_from_kernel: int | None = None
    for sched_idx, sched_item in list(enumerate(kernel.linearization))[::-1]:
        if isinstance(sched_item, CallKernel):
            return_from_kernel_idxs[sched_idx] = current_return_from_kernel
            current_return_from_kernel = None
        elif isinstance(sched_item, ReturnFromKernel):
            assert current_return_from_kernel is None
            current_return_from_kernel = sched_idx
            return_from_kernel_idxs[sched_idx] = current_return_from_kernel
        elif isinstance(sched_item, (RunInstruction, EnterLoop, LeaveLoop,
                                     Barrier)):
            return_from_kernel_idxs[sched_idx] = current_return_from_kernel
        else:
            raise NotImplementedError(type(sched_item))

    return return_from_kernel_idxs

# }}}


# {{{ check for write races in accesses

def _check_for_access_races(map_a, insn_a, map_b, insn_b, knl, callables_table,
                            address_space):
    """
    Returns *True* if the execution instances of *insn_a* and *insn_b*, accessing
    the same variable via access maps *map_a* and *map_b*, result in an access race.

    :arg address_space: An instance of :class:`loopy.kernel.data.AddressSpace`
        of the variable whose accesses are being checked for a race.

    .. note::

        The accesses ``map_a``, ``map_b`` lead to write races iff there exists 2
        *unequal* global ids that access the same address.
    """
    import pymbolic.primitives as p

    from loopy.kernel.data import (
        AddressSpace,
        HardwareConcurrentTag,
        filter_iname_tags_by_type,
    )
    from loopy.kernel.tools import get_hw_axis_base_for_codegen
    from loopy.symbolic import aff_from_expr, aff_to_expr, isl_set_from_expr

    assert address_space in [AddressSpace.LOCAL, AddressSpace.GLOBAL]

    gsize, lsize = knl.get_grid_size_upper_bounds(callables_table,
                                                  return_dict=True)

    # {{{ Step 1: Preprocess the maps

    # Step 1.1: Project out inames which are also map's dims, but does not form the
    #           insn's within_inames
    # Step 1.2: Perform any offsetting required to the hw axes iname terms
    # Step 1.3: Project out sequential inames in the access maps
    # Step 1.4: Rename the dims with their iname tags i.e. (g.i or l.i)
    # Step 1.5: Name the ith output dims as _lp_dim{i}

    updated_maps = []

    for (map_, insn) in [
            (map_a, insn_a),
            (map_b, insn_b)]:
        dims_not_to_project_out = ({iname
                                    for iname in insn.within_inames
                                    if knl.iname_tags_of_type(
                                        iname, HardwareConcurrentTag)}
                                   | knl.all_params())
        map_ = map_.project_out_except(sorted(dims_not_to_project_out),
                                       [isl.dim_type.in_,
                                        isl.dim_type.param,
                                        isl.dim_type.div,
                                        isl.dim_type.cst])

        for name, (dt, pos) in map_.get_var_dict().items():
            if dt == isl.dim_type.in_:
                tag, = filter_iname_tags_by_type(knl.inames[name].tags,
                                                 HardwareConcurrentTag)

                iname_lower_bound = get_hw_axis_base_for_codegen(knl, name)

                if not iname_lower_bound.plain_is_zero():
                    # Hardware inames with nonzero base have an offset applied in
                    # code generation:
                    # https://github.com/inducer/loopy/blob/4e0b1c7635afe1473c8636377f8e7ef6d78dfd46/loopy/codegen/loop.py#L293-L297
                    # https://github.com/inducer/loopy/issues/600#issuecomment-1104066735

                    map_ = map_.add_dims(isl.dim_type.out, 1)
                    map_ = map_.move_dims(
                        isl.dim_type.in_, pos+1,
                        isl.dim_type.out, map_.dim(isl.dim_type.out)-1,
                        1
                    )
                    map_ = map_.set_dim_name(isl.dim_type.in_, pos+1, name+"'")

                    lbound_offset_expr_aff = aff_from_expr(
                        map_.domain().space,
                        (p.Variable(name+"'")
                         + aff_to_expr(iname_lower_bound)
                         - p.Variable(name))
                    )
                    lbound_offset_as_domain = lbound_offset_expr_aff.zero_basic_set()
                    map_ = map_.intersect_domain(lbound_offset_as_domain)

                    map_ = map_.project_out(dt, pos, 1)
                    assert map_.get_dim_name(dt, pos) == name+"'"
                    map_ = map_.set_dim_name(dt, pos, name)

                map_ = map_.set_dim_name(dt, pos, str(tag))

        for i_l in lsize:
            if f"l.{i_l}" not in map_.get_var_dict():
                ndim = map_.dim(isl.dim_type.in_)
                map_ = map_.add_dims(isl.dim_type.in_, 1)
                map_ = map_.set_dim_name(isl.dim_type.in_, ndim, f"l.{i_l}")

        for i_g in gsize:
            if f"g.{i_g}" not in map_.get_var_dict():
                ndim = map_.dim(isl.dim_type.in_)
                map_ = map_.add_dims(isl.dim_type.in_, 1)
                map_ = map_.set_dim_name(isl.dim_type.in_, ndim, f"g.{i_g}")

        for pos in range(map_.dim(isl.dim_type.out)):
            map_ = map_.set_dim_name(isl.dim_type.out, pos, f"_lp_dim{pos}")

        updated_maps.append(map_)

    map_a, map_b = updated_maps

    # }}}

    # {{{ Step 2: rename all lid's, gid's in map_a to lid.A, gid.A

    for name, (dt, pos) in map_a.get_var_dict().items():
        if dt == isl.dim_type.in_:
            map_a = map_a.set_dim_name(dt, pos, name+".A")

    # }}}

    # {{{ Step 3: rename all lid's, gid's in map_b to lid.B, gid.B

    for name, (dt, pos) in map_b.get_var_dict().items():
        if dt == isl.dim_type.in_:
            map_b = map_b.set_dim_name(dt, pos, name+".B")

    # }}}

    # {{{ Step 4: make map_a, map_b ISL sets

    map_a, map_b = isl.align_two(map_a, map_b)
    map_a = map_a.move_dims(isl.dim_type.in_, map_a.dim(isl.dim_type.in_),
                            isl.dim_type.out, 0, map_a.dim(isl.dim_type.out))

    map_b = map_b.move_dims(isl.dim_type.in_, map_b.dim(isl.dim_type.in_),
                            isl.dim_type.out, 0, map_b.dim(isl.dim_type.out))
    set_a = map_a.domain()
    set_b = map_b.domain()

    # }}}

    assert set_a.get_space() == set_b.get_space()

    # {{{ Step 5: create the set any(l.i.A != l.i.B) OR any(g.i.A != g.i.B)

    space = set_a.space
    unequal_local_id_set = isl.Set.empty(set_a.get_space())
    unequal_group_id_set = isl.Set.empty(set_a.get_space())
    equal_group_id_set = isl.BasicSet.universe(set_a.get_space())

    for i_l in lsize:
        lid_a = p.Variable(f"l.{i_l}.A")
        lid_b = p.Variable(f"l.{i_l}.B")
        unequal_local_id_set |= (isl_set_from_expr(space,
                                                   p.Comparison(lid_a, "!=", lid_b))
                                 )

    for i_g in gsize:
        gid_a = p.Variable(f"g.{i_g}.A")
        gid_b = p.Variable(f"g.{i_g}.B")
        unequal_group_id_set |= (isl_set_from_expr(space,
                                                   p.Comparison(gid_a, "!=", gid_b))
                                 )
        equal_group_id_set &= (isl_set_from_expr(space,
                                                 p.Comparison(gid_a, "==", gid_b))
                               )

    # }}}

    if address_space == AddressSpace.GLOBAL:
        return not (set_a
                    & set_b
                    & (unequal_local_id_set
                       | unequal_group_id_set)
                    ).is_empty()
    else:
        return not (set_a
                    & set_b
                    & unequal_local_id_set
                    & equal_group_id_set).is_empty()


class AccessMapDescriptor(enum.Enum):
    """
    Special access map values.

    :attr DOES_NOT_ACCESS: Describes an unaccessed variable.
    :attr NON_AFFINE_ACCESS: Describes a non-quasi-affine access into an array.
    """
    DOES_NOT_ACCESS = enum.auto()
    NON_AFFINE_ACCESS = enum.auto()


class WriteRaceChecker:
    """Used for checking for overlap between access ranges of instructions."""

    def __init__(self, kernel, callables_table):
        self.kernel = kernel
        self.callables_table = callables_table

    @cached_property
    def vars(self):
        return (self.kernel.get_written_variables()
                | self.kernel.get_read_variables())

    @memoize_method
    def _get_access_maps(self, insn_id, access_dir):
        from collections import defaultdict

        from loopy.symbolic import BatchedAccessMapMapper

        insn = self.kernel.id_to_insn[insn_id]

        exprs = list(insn.assignees)
        if access_dir == "any":
            exprs.append(insn.expression)
            exprs.extend(insn.predicates)

        access_maps = defaultdict(lambda: AccessMapDescriptor.DOES_NOT_ACCESS)

        arm = BatchedAccessMapMapper(self.kernel, self.vars, overestimate=True)

        for expr in exprs:
            arm(expr, insn.within_inames)

        for name in arm.access_maps:
            if arm.bad_subscripts[name]:
                access_maps[name] = AccessMapDescriptor.NON_AFFINE_ACCESS
                continue
            access_maps[name] = arm.access_maps[name][insn.within_inames]

        return access_maps

    def _get_access_map_for_var(self, insn_id, access_dir, var_name):
        assert access_dir in ["w", "any"]

        insn = self.kernel.id_to_insn[insn_id]
        # Access range checks only apply to assignment-style instructions. For
        # non-assignments, we rely on read/write dependency information.
        from loopy.kernel.instruction import MultiAssignmentBase
        if not isinstance(insn, MultiAssignmentBase):
            if access_dir == "any":
                return var_name in insn.dependency_names()
            else:
                return var_name in insn.write_dependency_names()

        return self._get_access_maps(insn_id, access_dir)[var_name]

    def do_accesses_result_in_races(self, insn1, insn1_dir, insn2, insn2_dir,
                                    var_name):
        """Determine whether the access maps to *var_name* in the two given
        instructions result in write races owing to concurrent iname tags. This
        determination is made 'conservatively', i.e. if precise information is
        unavailable (for ex. if one of the instructions accesses *var_name* via
        indirection), it is concluded that the ranges overlap.

        :arg insn1_dir: either ``"w"`` or ``"any"``, to indicate which
            type of access is desired--writing or any
        :arg insn2_dir: either ``"w"`` or ``"any"``
        :returns: a :class:`bool`
        """

        insn1_amap = self._get_access_map_for_var(insn1, insn1_dir, var_name)
        insn2_amap = self._get_access_map_for_var(insn2, insn2_dir, var_name)

        if (insn1_amap is AccessMapDescriptor.DOES_NOT_ACCESS
                or insn2_amap is AccessMapDescriptor.DOES_NOT_ACCESS):
            return False
        if (insn1_amap is AccessMapDescriptor.NON_AFFINE_ACCESS
                or insn2_amap is AccessMapDescriptor.NON_AFFINE_ACCESS):
            return True

        return _check_for_access_races(insn1_amap, self.kernel.id_to_insn[insn1],
                                       insn2_amap, self.kernel.id_to_insn[insn2],
                                       self.kernel, self.callables_table,
                                       (self.kernel
                                        .get_var_descriptor(var_name)
                                        .address_space))

# }}}


LoopNestTree: TypeAlias = Tree[InameStrSet]
LoopTree: TypeAlias = Tree[InameStr]


class V2SchedulerNotImplementedError(LoopyError):
    pass


def separate_loop_nest(
            tree: LoopNestTree,
            loop_nests: Collection[InameStrSet],
            inames_to_separate: InameStrSet
        ) -> tuple[LoopNestTree, InameStrSet, InameStrSet | None]:
    """
    Returns a copy of *tree* that has *inames_to_separate* occur in
    nodes that are not shared with other inames.
    Returns a version of the loop nest tree *tree* so that every node in the tree is
    either a subset of *outermost_inames* or has an empty intersection with
    *outermost_inames*.

    This routine modifies at most one node of the tree.
    All its ancestors must satisfy `ancestor <= outermost_inames`.
    For the first node not  satisfying this relationship,
    if `node & outermost_inames` is empty, no modification is made.
    Otherwise, if ``node & outermost_inames < node``, that node is split
    so as to separate *outermost_inames* in their own node.

    :arg loop_nests: A collection of nodes in *tree* that cover
        *inames_to_separate*.

    :returns: a :class:`tuple` ``(new_tree, outer_loop_nest, inner_loop_nest)``,
        where outer_loop_nest is the identifier for the new outer and inner
        loop nests so that *inames_to_separate* is a valid nesting.

    .. note::

        We could compute *loop_nests* within this routine's implementation, but
        computing would be expensive and hence we ask the caller for this info.

    Example::
       *tree*: frozenset()
               └── frozenset({'j', 'i'})
                   └── frozenset({'k', 'l'})

       *inames_to_separate*: frozenset({'k', 'i', 'j'})
       *loop_nests*: {frozenset({'j', 'i'}), frozenset({'k', 'l'})}

       Returns:

       *new_tree*: frozenset()
                   └── frozenset({'j', 'i'})
                       └── frozenset({'k'})
                           └── frozenset({'l'})

       *outer_loop_nest*: frozenset({'k'})
       *inner_loop_nest*: frozenset({'l'})
    """
    assert all(isinstance(loop_nest, frozenset) for loop_nest in loop_nests)

    # annotation to avoid https://github.com/python/mypy/issues/17693
    emptyset: InameStrSet = frozenset()

    assert inames_to_separate <= reduce(frozenset.union, loop_nests, emptyset)

    # {{{ sanity check to ensure the loop nest *inames_to_separate* is possible

    loop_nests = sorted(loop_nests, key=lambda nest: tree.depth(nest))

    for outer, inner in itertools.pairwise(loop_nests):
        if outer != tree.parent(inner):
            raise LoopyError(f"Cannot schedule loop nest {inames_to_separate} "
                             f" in the nesting tree:\n{tree}")

    assert tree.depth(loop_nests[0]) == 0

    # }}}

    innermost_node = loop_nests[-1]
    # separate variable to avoid https://github.com/python/mypy/issues/17694
    outerer_loops = reduce(frozenset.union, loop_nests[:-1], emptyset)
    new_outer_node = inames_to_separate - outerer_loops
    new_inner_node = innermost_node - inames_to_separate

    if new_outer_node == innermost_node:
        # such a loop nesting already exists => do nothing
        return tree, new_outer_node, None

    # add the outer loop to our loop nest tree
    tree = tree.add_node(new_outer_node,
                         parent=not_none(tree.parent(innermost_node)))

    # rename the old loop to the inner loop
    tree = tree.replace_node(innermost_node,
                            new_node=new_inner_node)

    # set the parent of inner loop to be the outer loop
    tree = tree.move_node(new_inner_node, new_parent=new_outer_node)

    return tree, new_outer_node, new_inner_node


def _add_inner_loops(tree, outer_loop_nest, inner_loop_nest):
    """
    Returns a copy of *tree* that nests *inner_loop_nest* inside *outer_loop_nest*.
    """
    # add the outer loop to our loop nest tree
    return tree.add_node(inner_loop_nest, parent=outer_loop_nest)


def _order_loop_nests(
            loop_nest_tree: LoopNestTree,
            strict_priorities: frozenset[tuple[InameStr, ...]],
            relaxed_priorities: frozenset[tuple[InameStr, ...]],
            iname_to_tree_node_id: Mapping[InameStr, InameStrSet],
          ) -> LoopTree:
    """
    Returns a loop nest where all nodes in the tree are instances of
    :class:`str` denoting inames. Unlike *loop_nest_tree* which corresponds to
    multiple loop nesting, this routine returns a unique loop nest that is
    obtained after constraining *loop_nest_tree* with the constraints enforced
    by *priorities*.

    :arg strict_priorities: Expresses strict nesting constraints using the same
        data structure as :attr:`loopy.LoopKernel.loop_priority`.
        These priorities are imposed strictly i.e. if these conditions cannot be met a
        :class:`loopy.diagnostic.LoopyError` is raised.

    :arg relaxed_priorities: Expresses strict nesting constraints using the same
        data structure as :attr:`loopy.LoopKernel.loop_priority`.
        These nesting constraints are treated as optional.

    :arg iname_to_tree_node_id: A mapping from iname to the loop nesting its a
        part of.
    """
    from warnings import warn

    from pytools.graph import compute_topological_order as toposort

    loop_nests = set(iname_to_tree_node_id.values())

    # nesting_constraints: A mapping from the loop nest level to the nesting
    # constraints applicable to it.
    # Each nesting constraint is represented as a DAG. In the DAG, if there
    # exists an edge from from iname 'i' -> iname 'j' => 'j' should be nested
    # inside 'i'.
    iname_to_nesting_constraints: dict[InameStrSet, dict[InameStr, InameStrSet]] = {
        loop_nest: {iname: frozenset() for iname in loop_nest}
        for loop_nest in loop_nests}

    # The plan here is populate DAGs in *nesting_constraints* and then perform a
    # toposort for each loop nest.

    def _update_nesting_constraints(
                priorities: frozenset[tuple[InameStr, ...]],
                cannot_satisfy_callback: Callable[[str], None]
            ) -> None:
        """
        Records *priorities* in *nesting_constraints* and calls
        *cannot_satisfy_callback* with an appropriate error message if the
        priorities cannot be met.
        """
        for priority in priorities:
            for outer_iname, inner_iname in itertools.pairwise(priority):
                if inner_iname not in iname_to_tree_node_id:
                    cannot_satisfy_callback(f"Cannot enforce the constraint:"
                                            f" {inner_iname} to be nested within"
                                            f" {outer_iname}, as {inner_iname}"
                                            f" is either a parallel loop or"
                                            f" not an iname.")
                    continue

                if outer_iname not in iname_to_tree_node_id:
                    cannot_satisfy_callback(f"Cannot enforce the constraint:"
                                            f" {inner_iname} to be nested within"
                                            f" {outer_iname}, as {outer_iname}"
                                            f" is either a parallel loop or"
                                            f" not an iname.")
                    continue

                inner_iname_nest = iname_to_tree_node_id[inner_iname]
                outer_iname_nest = iname_to_tree_node_id[outer_iname]

                if inner_iname_nest == outer_iname_nest:
                    iname_to_nesting_constraints[
                        inner_iname_nest][outer_iname] |= {inner_iname}
                else:
                    ancestors_of_inner_iname = (loop_nest_tree
                                                .ancestors(inner_iname_nest))
                    ancestors_of_outer_iname = (loop_nest_tree
                                                .ancestors(outer_iname_nest))
                    if any(outer_iname in ancestor
                           for ancestor in ancestors_of_inner_iname):
                        # nesting constraint already satisfied => do nothing
                        pass
                    elif any(inner_iname in ancestor
                             for ancestor in ancestors_of_outer_iname):
                        cannot_satisfy_callback("Cannot satisfy constraint that"
                                                f" iname '{inner_iname}' must be"
                                                f" nested within '{outer_iname}''.")
                    else:
                        # inner iname and outer iname are indirect family members
                        # => must be realized via dependencies in the linearization
                        # phase, not implemented in v2-scheduler yet.
                        raise V2SchedulerNotImplementedError("cannot"
                                " schedule kernels with priority dependencies"
                                " between sibling loop nests")

    def _raise_loopy_err(x):
        raise LoopyError(x)

    # record strict priorities
    _update_nesting_constraints(strict_priorities, _raise_loopy_err)
    # record relaxed priorities
    _update_nesting_constraints(relaxed_priorities, warn)

    # ordered_loop_nests: A mapping from the unordered loop nests to their
    # ordered counterparts. For example. If we had only one loop nest
    # `frozenset({"i", "j", "k"})`, and the prioirities said added the
    # constraint that "i" must be nested within "k", then `ordered_loop_nests`
    # would be: `{frozenset({"i", "j", "k"}): ["j", "k", "i"]}` i.e. the loop
    # nests would now have an order.
    ordered_loop_nests = {
        unordered_nest: toposort(flow, key=lambda x: x)
        for unordered_nest, flow in iname_to_nesting_constraints.items()}

    # {{{ combine 'loop_nest_tree' along with 'ordered_loop_nest_tree'

    assert loop_nest_tree.root == frozenset()

    new_tree = Tree.from_root("")

    old_to_new_parent = {}

    old_to_new_parent[loop_nest_tree.root] = ""

    # traversing 'tree' in an BFS fashion to create 'new_tree'
    queue = list(loop_nest_tree.children(loop_nest_tree.root))

    while queue:
        current_nest = queue.pop(0)

        ordered_nest = ordered_loop_nests[current_nest]
        new_tree = new_tree.add_node(ordered_nest[0],
                                     parent=old_to_new_parent[not_none(loop_nest_tree
                                                              .parent(current_nest))])
        for new_parent, new_child in itertools.pairwise(ordered_nest):
            new_tree = new_tree.add_node(node=new_child, parent=new_parent)

        old_to_new_parent[current_nest] = ordered_nest[-1]

        queue.extend(loop_nest_tree.children(current_nest))

    # }}}

    return new_tree


@memoize_on_first_arg
def _get_parallel_inames(kernel: LoopKernel) -> AbstractSet[str]:
    from loopy.kernel.data import ConcurrentTag, IlpBaseTag, VectorizeTag

    concurrent_inames = {iname for iname in kernel.all_inames()
                         if kernel.iname_tags_of_type(iname, ConcurrentTag)}
    ilp_inames = {iname for iname in kernel.all_inames()
                  if kernel.iname_tags_of_type(iname, IlpBaseTag)}
    vec_inames = {iname for iname in kernel.all_inames()
                  if kernel.iname_tags_of_type(iname, VectorizeTag)}
    return (concurrent_inames - ilp_inames - vec_inames)


def get_partial_loop_nest_tree(kernel: LoopKernel) -> LoopNestTree:
    """
    Returns a tree representing the *kernel*'s loop nests.

    Each node of the returned tree has a :class:`frozenset` of inames.
    All the inames in the identifier of a parent node of a loop nest in the
    tree must be nested outside all the iname in identifier of the loop nest.

    .. note::

        This routine only takes into account the nesting dependency
        constraints of :attr:`loopy.InstructionBase.within_inames` of all the
        *kernel*'s instructions and the iname tags. This routine does *NOT*
        include the nesting constraints imposed by the dependencies between the
        instructions and the dependencies imposed by the kernel's domain tree.
    """
    from loopy.kernel.data import IlpBaseTag

    # figuring the possible loop nestings minus the concurrent_inames as they
    # are never realized as actual loops
    insn_iname_sets = {
        insn.within_inames - _get_parallel_inames(kernel)
        for insn in kernel.instructions}

    root: InameStrSet = frozenset()
    tree = Tree.from_root(root)

    # mapping from iname to the innermost loop nest they are part of in *tree*.
    iname_to_tree_node_id: dict[InameStr, InameStrSet] = {}

    # if there were any loop with no inames, those have been already account
    # for as the root.
    insn_iname_sets = insn_iname_sets - {root}

    for iname_set in insn_iname_sets:
        not_seen_inames = frozenset(iname for iname in iname_set
                                    if iname not in iname_to_tree_node_id)
        seen_inames = iname_set - not_seen_inames

        all_nests = {iname_to_tree_node_id[iname] for iname in seen_inames}

        tree, outer_loop, inner_loop = separate_loop_nest(tree,
                                                           (all_nests
                                                            | {frozenset()}),
                                                           seen_inames)
        if not_seen_inames:
            # make '_not_seen_inames' nest inside the seen ones.
            # example: if there is already a loop nesting "i,j,k"
            # and the current iname chain is "i,j,l". Only way this is possible
            # is if "l" is nested within "i,j"-loops.
            tree = _add_inner_loops(tree, outer_loop, not_seen_inames)

        # {{{ update iname to node id

        for iname in outer_loop:
            iname_to_tree_node_id[iname] = outer_loop

        if inner_loop is not None:
            for iname in inner_loop:
                iname_to_tree_node_id[iname] = inner_loop

        for iname in not_seen_inames:
            iname_to_tree_node_id[iname] = not_seen_inames

        # }}}

    # {{{ make ILP tagged inames innermost

    ilp_inames = {iname for iname in kernel.all_inames()
                  if kernel.iname_tags_of_type(iname, IlpBaseTag)}

    for iname_set in insn_iname_sets:
        for ilp_iname in (ilp_inames & insn_iname_sets):
            # pull out other loops so that ilp_iname is the innermost
            all_nests = {iname_to_tree_node_id[iname] for iname in seen_inames}
            tree, outer_loop, inner_loop = separate_loop_nest(tree,
                                                               (all_nests
                                                                | {frozenset()}),
                                                               (iname_set
                                                                - {ilp_iname}))

            for iname in outer_loop:
                iname_to_tree_node_id[iname] = outer_loop

            if inner_loop is not None:
                for iname in inner_loop:
                    iname_to_tree_node_id[iname] = inner_loop

    # }}}

    return tree


def _get_iname_to_tree_node_id_from_partial_loop_nest_tree(
            tree: LoopNestTree,
        ) -> Mapping[str, frozenset[str]]:
    """
    Returns the mapping from the iname to the *tree*'s node that it was a part
    of.

    :arg tree: A partial loop nest tree.
    """
    iname_to_tree_node_id = {}
    for node in tree.nodes():
        assert isinstance(node, frozenset)
        for iname in node:
            iname_to_tree_node_id[iname] = node

    return constantdict(iname_to_tree_node_id)


def get_loop_tree(kernel: LoopKernel) -> LoopTree:
    """
    Returns a tree representing the loop nesting for *kernel*. A parent node in
    the tree is always nested outside all its children.

    .. note::

        Multiple loop nestings might exist for *kernel*, but this routine returns
        one valid loop nesting.
    """
    from islpy import dim_type

    tree = get_partial_loop_nest_tree(kernel)
    iname_to_tree_node_id = (
        _get_iname_to_tree_node_id_from_partial_loop_nest_tree(tree))

    strict_loop_priorities: frozenset[tuple[InameStr, ...]] = frozenset()

    # {{{ impose constraints by the domain tree

    # FIXME: These three could be one statement if it weren't for
    # - https://github.com/python/mypy/issues/17693
    # - https://github.com/python/mypy/issues/17694
    emptyset: InameStrSet = frozenset()
    loop_inames = reduce(frozenset.union,
                          (insn.within_inames
                           for insn in kernel.instructions),
                          emptyset)
    loop_inames = loop_inames - _get_parallel_inames(kernel)

    for dom in kernel.domains:
        for outer_iname in set(dom.get_var_names(dim_type.param)):
            if outer_iname not in loop_inames:
                continue

            for inner_iname in dom.get_var_names(dim_type.set):
                if inner_iname not in loop_inames:
                    continue

                # either outer_iname and inner_iname should belong to the same
                # loop nest level or outer should be strictly outside inner
                # iname
                inner_iname_nest = iname_to_tree_node_id[inner_iname]
                outer_iname_nest = iname_to_tree_node_id[outer_iname]

                if inner_iname_nest == outer_iname_nest:
                    strict_loop_priorities |= {(outer_iname, inner_iname)}
                else:
                    ancestors_of_inner_iname = tree.ancestors(inner_iname_nest)
                    if outer_iname_nest not in ancestors_of_inner_iname:
                        raise LoopyError(f"Loop '{outer_iname}' cannot be nested"
                                         f" outside '{inner_iname}'.")

    # }}}

    return _order_loop_nests(tree,
                             strict_loop_priorities,
                             kernel.loop_priority,
                             iname_to_tree_node_id)

# vim: fdm=marker
